from pathlib import Path

from mist.app.dbs import DOWNLOADERS


class MistDownload:
    """
    Wrapper for downloading schemes.
    """

    def __init__(self, url: str, output: Path, include_profiles: bool, downloader: str, dir_tokens: Path, key_name: str,
                 site: str) -> None:
        """
        Initialize the scheme download class.
        :param url: Scheme URL
        :param output: Output directory
        :param include_profiles: If True, profiles are kept
        :param downloader: Key of the downloader class
        :param dir_tokens: Token directory
        :param key_name: Key name
        :param site: Site name
        :return: None
        """
        self._url = url
        self._output = output
        self._include_profiles = include_profiles
        self._downloader = downloader
        self._dir_tokens = dir_tokens
        self._key_name = key_name
        self._site = site

    def run(self) -> None:
        """
        Runs the downloader.
        :return: None
        """
        self._output.mkdir(exist_ok=True, parents=True)
        downloader = DOWNLOADERS[self._downloader](
            dir_tokens=self._dir_tokens,
            key_name=self._key_name,
            site=self._site,
        )
        downloader.download_scheme(self._url, self._output, self._include_profiles)



from pathlib import Path

from Bio import SeqIO

from mist.app.loggers.logger import logger
from mist.app.utils import sequenceutils


class SeqHolder:
    """
    Class to store and retrieve sequences from the input FASTA file.
    """

    def __init__(self, path_fasta: Path | None = None, seqs: list[str] | None = None) -> None:
        """
        Initializes the holder.
        :param path_fasta: Input FASTA path
        :param seqs: List of sequences
        :return: None
        """
        if path_fasta is not None:
            with path_fasta.open() as handle:
                self._seq_by_id = {s.id: str(s.seq) for s in SeqIO.parse(handle, 'fasta')}
        else:
            self._seq_by_id = {str(i): seq for i, seq in enumerate(seqs)}
        logger.debug(f'Parsed {len(self._seq_by_id):,} sequence(s)')

    def get_seq(self, seq_id: str, start: int, end: int, strand: str | None = None) -> str:
        """
        Retrieves the target sequence.
        :param seq_id: Sequence id
        :param start: Sequence start position
        :param end: Sequence end position
        :param strand: Strand (+ / -)
        :return: Sequence
        """
        if strand in ('+', None):
            return self._seq_by_id[seq_id][start-1:end]
        return sequenceutils.rev_complement(self._seq_by_id[seq_id][start-1:end])



#!/usr/bin/env python
from datetime import datetime
from importlib.resources import files
from pathlib import Path
from typing import Callable

import click

from mist.app.dbs import DOWNLOADERS
from mist.app.loggers.logger import initialize_logging, logger
from mist.app.query.allelequeryminimap import MultiStrategy
from mist.scripts.mistcaller import MistCaller
from mist.scripts.mistdists import MistDists
from mist.scripts.mistdownload import MistDownload
from mist.scripts.mistindex import MistIndex
from mist.scripts.mistlist import MistList
from mist.version import __version__


def _common_options(func: Callable) -> Callable:
    """
    Defines the common options.
    :param func: Function to decorate
    :return: Decorated function
    """
    func = click.option("--log", type=Path, help="Save log to this file")(func)
    func = click.option("--debug", is_flag=True, help="Enable debug mode")(func)
    return func


@click.group(help='MiST: Minimap2-inferred Sequence Typing.')
@click.version_option(__version__, prog_name="MiST", message="%(prog)s %(version)s")
def cli() -> None:
    """
    MiST: Minimap2-inferred Sequence Typing.
    """
    pass

@cli.command(name='index')
@click.argument("fasta", nargs=-1, type=click.Path(exists=True))
@click.option("-l", "--fasta-list", type=click.Path(path_type=Path, exists=True), help="List with input FASTA path(s)")
@click.option("-p", "--profiles", type=click.Path(path_type=Path, exists=True), help="TSV file with profiles")
@click.option("-o", "--output", type=click.Path(path_type=Path), required=True, help="Output directory")
@click.option("-c", "--cutoff", type=int, default=95, show_default=True, help="Clustering cutoff")
@click.option("-t", "--threads", type=int, default=1, show_default=True, help="Number of threads to use")
@_common_options
def index_(
        fasta: list[Path],
        fasta_list: Path,
        profiles: Path | None,
        output: Path,
        cutoff: int,
        threads: int,
        debug: bool,
        log: Path) -> None:
    """
    Creates a MIST index for a set of input FASTA files.
    """
    initialize_logging(log_path=log, debug=debug)

    # Collect FASTA files
    if len(fasta) > 0:
        paths_fasta = [Path(f) for f in fasta]
    elif fasta_list is not None:
        with open(fasta_list) as handle:
            paths_fasta = [Path(line.strip()) for line in handle.readlines() if len(line.strip()) > 0]
    else:
        raise click.UsageError("No FASTA file or FASTA list provided.")
    if len(paths_fasta) < 1:
        raise click.UsageError("No input FASTA file(s) provided.")

    # Run the indexer
    indexer = MistIndex(
        paths_fasta=paths_fasta,
        path_profiles=profiles,
        cutoff=cutoff,
        debug=debug
    )
    indexer.create_index(dir_out=output.expanduser().resolve(), threads=threads)


@cli.command()
@click.option("-f", "--fasta", type=click.Path(exists=True, path_type=Path, dir_okay=False), required=True, help="Input FASTA path")
@click.option("-d", "--db", type=click.Path(exists=True, path_type=Path), required=True, help="Database path")
@click.option("-o", "--out-json", type=click.Path(), default=None, help="JSON output file")
@click.option("-t", "--threads", type=int, default=1, show_default=True, help="Number of threads to use")
@click.option("--out-tsv", type=click.Path(path_type=Path, writable=True), help="TSV output file")
@click.option("--out-dir", type=click.Path(path_type=Path), help="Output directory")
@click.option("--export-novel", is_flag=True, help="Create FASTA files for (potential) novel alleles")
@click.option("--keep-minimap2", is_flag=True, help="Store the minimap2 output")
@click.option("--min-id-novel", type=int, default=99, show_default=True, help="Minimum % identity for novel alleles")
@click.option(
    "-m", "--multi", type=click.Choice([s.value for s in MultiStrategy]), default=MultiStrategy.ALL.value,
    show_default=True, help="Strategy to handle multiple perfect hits")
@click.option("--loci", help="Limit to these loci, provided as comma seperated string (e.g., 'abcZ,fumC')")
@_common_options
def call(
        fasta: Path,
        db: Path,
        out_json: Path,
        out_tsv: Path,
        out_dir: Path,
        export_novel: bool,
        keep_minimap2: bool,
        threads: int,
        min_id_novel: int,
        multi: str,
        loci: str | None,
        debug: bool,
        log: Path) -> None:
    """
    Calls alleles from a FASTA file.
    """
    # Setup
    initialize_logging(log_path=log, debug=debug)
    t0 = datetime.now()

    # Validate args
    if (export_novel or keep_minimap2) and (out_dir is None):
        raise click.UsageError("Output directory ('--out-dir') must be specified when exporting FASTA / Minimap2 files")

    # Create the output directory
    if out_dir is not None:
        out_dir.mkdir(parents=True, exist_ok=True)

    # Update the JSON output path
    if (out_json is None) and (out_dir is not None):
        out_json_ = Path(out_dir, 'mist.json')
    elif out_json is not None:
        out_json_ = out_json
    else:
        out_json_ = Path('mist.json')

    # Call alleles
    caller = MistCaller(
        dir_db=db,
        multi=multi,
        keep_minimap2=keep_minimap2,
        loci=loci.split(',') if loci else None,
        export_novel=export_novel,
        min_id_novel=min_id_novel
    )
    caller.call_alleles(
        path_fasta=fasta,
        out_json=out_json_,
        out_tsv=out_tsv,
        out_dir=out_dir,
        threads=threads
    )
    logger.info("Make sure to cite the corresponding database when using this in your research")
    path_citation = files('mist').joinpath('resources/citation.txt')
    logger.info(f'Please cite: {path_citation.read_text()}')
    logger.info(f"Processing time: {(datetime.now() - t0).total_seconds():.2f} seconds")


@cli.command()
@click.option("--url", required=True, help="URL to download from.")
@click.option("-o", "--output", type=click.Path(path_type=Path), default=Path("mist_download"), show_default=True, help="Output directory")
@click.option("-p", "--include-profiles", is_flag=True, help="Download the profiles")
@click.option("-d", "--downloader", type=click.Choice(list(DOWNLOADERS.keys())), required=True, help="Downloader")
@click.option("--dir-tokens", type=click.Path(path_type=Path), default=Path.cwd() / ".bigsdb_tokens", show_default=True, help="Directory with access tokens")
@click.option("--key-name", default="PubMLST", show_default=True, help="Key name")
@click.option("--site", default="PubMLST", show_default=True, help="Site")
@_common_options
def download(
        url: str,
        output: Path,
        include_profiles: bool,
        downloader: str,
        dir_tokens: Path,
        key_name: str,
        site: str,
        debug: bool,
        log: Path) -> None:
    """
    Download (cg)MLST schemes from various sources.
    """
    initialize_logging(log_path=log, debug=debug)
    downloader = MistDownload(
        url=url,
        output=output,
        include_profiles=include_profiles,
        downloader=downloader,
        dir_tokens=dir_tokens,
        key_name=key_name,
        site=site
    )
    downloader.run()

@cli.command()
@click.argument('inputs', nargs=-1)
@click.option("-d", "--out-dists", type=click.Path(path_type=Path), default=Path("distances.tsv"), show_default=True, help="Distance matrix output")
@click.option("-m", "--out-matrix", type=click.Path(path_type=Path), default=Path("allele_matrix.tsv"), show_default=True, help="Allele matrix output")
@click.option("-l", "--min-perc-loci", type=int, default=90, show_default=True, help="Minimum percentage of loci that should be present in a dataset")
@click.option("-s", "--min-perc-samples", type=int, default=90, show_default=True, help="Minimum percentage of datasets where loci should be present")
@_common_options
def dists(
        inputs: tuple[Path],
        out_dists: Path,
        out_matrix: Path,
        min_perc_loci: int,
        min_perc_samples: int,
        debug: bool,
        log: Path) -> None:
    """
    Builds distance and allele matrices from MiST output files.
    """
    initialize_logging(log_path=log, debug=debug)
    if len(inputs) == 0:
        raise click.UsageError("No input files provided.")

    mist_dists = MistDists(
        inputs=[Path(x) for x in inputs],
        out_matrix=out_matrix,
        out_dists=out_dists,
        min_perc_loci=min_perc_loci,
        min_perc_samples=min_perc_samples
    )
    mist_dists.run()

@cli.command(name='list')
@click.option('-d', '--downloader', type=click.Choice(list(DOWNLOADERS.keys())), required=True, help="Downloader")
@click.option('-h', '--host', help="Host name (required only if multiple hosts are available for this downloader).")
@click.option('--db', type=str, help="BIGSdb only: database name")
def list_(downloader: str, host: str | None, db: str) -> None:
    """
    Lists the available schemes that can be downloaded.
    """
    # Validate the arguments
    if downloader in ('bigsdb', 'bigsdb_auth') and host is None:
        raise click.UsageError("--host option is required for this downloader. Valid options: 'pubmlst' or 'pasteur'.")

    # Retrieve available schemes
    mist_list = MistList(
        downloader=downloader,
        host=host,
        db=db
    )
    mist_list.print_available_schemes()


if __name__ == "__main__":
    cli()



import hashlib
import re
from pathlib import Path

import pandas as pd
from Bio import SeqIO
from Bio.Seq import Seq

REGEX_ALLELE = r'_(\d+)$'


def rev_complement(seq_in: str) -> str:
    """
    Returns the reverse complement of the input sequence.
    :param seq_in: Input sequence
    :return: Reverse complement
    """
    return str(Seq(seq_in).reverse_complement())


def hash_sequence(seq_in: str, rev_comp: bool = False) -> str:
    """
    Hashes the input sequence.
    :param seq_in: Input sequence
    :param rev_comp: Reverse complement the sequence before hashing
    :return: Hashed sequence
    """
    seq_in = seq_in.lower().replace('-', '')
    if rev_comp is True:
        seq_in = rev_complement(seq_in)
    return hashlib.sha1(seq_in.encode('ascii')).hexdigest()


def extract_hashes(path_fasta: Path, ori_by_seq_id: dict[str, str]) -> pd.DataFrame:
    """
    Extracts the hashes from the target FASTA file.
    :param path_fasta: Input FASTA file
    :param ori_by_seq_id: The sequence orientation by sequence id
    :return: DataFrame with hashes
    """
    records_out = []
    with path_fasta.open() as handle:
        for record in SeqIO.parse(handle, 'fasta'):
            if ori_by_seq_id[record.id] in ('+', None):
                seq = str(record.seq)
            else:
                seq = rev_complement(record.seq)
            records_out.append({
                'seq_id': record.id,
                'hash': hash_sequence(seq),
            })
    data_out = pd.DataFrame(records_out)
    data_out['allele_id'] = data_out['seq_id'].apply(lambda x: re.search(REGEX_ALLELE, x).group(1))
    return data_out


def merge_fasta_files(paths_fasta: list[Path], path_out: Path) -> int:
    """
    Creates a FASTA file that combines all representative alleles.
    :param paths_fasta: Input FASTA files
    :param path_out: Output path
    :return: Number of sequences
    """
    sequences_added = 0
    with open(path_out, 'w') as handle_out:
        for path_fasta in paths_fasta:
            with path_fasta.open() as handle_in:
                for seq in SeqIO.parse(handle_in, 'fasta'):
                    SeqIO.write(seq, handle_out, 'fasta')
                    sequences_added += 1
    return sequences_added


def count_sequences(path_in: Path) -> int:
    """
    Counts the number of sequences in the input FASTA file.
    :param path_in: Input path
    :return: Number of sequences
    """
    with path_in.open() as handle:
        return len(list(SeqIO.parse(handle, 'fasta')))



import json
from enum import Enum
from pathlib import Path
from typing import Optional

import pandas as pd

from mist.app import model
from mist.app.loggers.logger import logger
from mist.app.query.bestmatching import ImperfectMatchDetector, InvalidLengthException
from mist.app.query.seqholder import SeqHolder
from mist.app.utils import (
    minimap2utils,
    sequenceutils,
    unique_preserve_order,
)


class MultiStrategy(Enum):
    """
    Strategy to handle multiple perfect hits.
    """
    ALL = 'all'
    FIRST = 'first'
    LONGEST = 'longest'


def merge_results(results: list[model.AlleleResult], multi: str) -> str:
    """
    Merges allele results into a single allele string.
    :param results: Results
    :param multi: Multi-hit strategy
    :return: Allele string
    """
    if len(results) == 0:
        return '-'
    if len(results) == 1:
        return results[0].allele
    multi_hit_strategy = MultiStrategy(multi)
    if multi_hit_strategy == MultiStrategy.ALL:
        unique_alleles = unique_preserve_order([r.allele for r in results])
        return '__'.join(unique_alleles)
    elif multi_hit_strategy == MultiStrategy.LONGEST:
        max_length = max(res.alignment.length for res in results)
        unique_alleles = unique_preserve_order([r.allele for r in results if r.alignment.length == max_length])
        return '__'.join(unique_alleles)
    else:
        raise ValueError(f'Invalid multi-hit strategy: {multi}')


class AlleleQueryMinimap2:
    """
    Queries alleles using Minimap2.
    """

    def __init__(
            self,
            dir_db: Path,
            dir_out: Optional[Path] = None,
            multi_strategy: MultiStrategy = MultiStrategy.LONGEST,
            min_id_novel: int = 99,
            save_minimap2: bool = False
    ) -> None:
        """
        Initializes this class.
        :param dir_db: Path to the database
        :param dir_out: Output directory to store BLAST output (optional)
        :param multi_strategy: Strategy to handle multiple perfect hits
        :param min_id_novel: Minimum % identity for novel alleles
        :param save_minimap2: If True, the Minimap2 output is stored
        :return: None
        """
        self._dir_db = dir_db
        self._dir_out = dir_out
        self._multi_strategy = multi_strategy
        self._min_id_novel = min_id_novel
        self._seq_holder: SeqHolder | None = None
        self._save_minimap2 = save_minimap2

    @staticmethod
    def __get_coord(x: pd.Series) -> tuple[str, int, int]:
        """
        Constructs the coordinate string to retrieve the target sequence from the input assembly.
        :param x: BLAST output records
        :return: Coordinate string
        """
        if x['sstrand'] in ('plus', '+'):
            overhang_left = x['sstart']
            overhang_right = x['slen'] - x['send']
            return x['qseqid'], x['qstart'] + 1 - overhang_left, x['qend'] + overhang_right
        else:
            overhang_left = x['sstart']
            overhang_right = x['slen'] - x['send']
            return x['qseqid'], x['qstart'] + 1 - overhang_right, x['qend'] + overhang_left

    def _extract_exact_match(self, seq: str, row: pd.Series, data_locus: pd.DataFrame) -> model.AlleleResult | None:
        """
        Checks for an exact match.
        :param seq: Allele sequence
        :param row: Alignment record
        :param data_locus: Locus data
        :return: Match
        """
        allele = data_locus['hashes'].get(sequenceutils.hash_sequence(seq))
        if allele is None:
            allele = data_locus['hashes'].get(sequenceutils.hash_sequence(sequenceutils.rev_complement(seq)))
        if allele is None:
            return None
        return model.AlleleResult(
            allele=allele,
            alignment=model.Alignment(
                seq_id=row['query_name'],
                start=row['query_start'],
                end=row['query_end'],
                strand=row['sstrand']),
            length=len(seq)
        )

    def _extract_partial_match(self, df_alignment: pd.DataFrame, locus_name: str) -> model.QueryResult | None:
        """
        Checks for an exact match.
        :param df_alignment: Minimap2 alignment data
        :param locus_name: Locus name
        :return: Match
        """
        divergence = df_alignment['tag_dv'].str.rsplit(':', n=1).str[-1].astype(float)
        best_idx = divergence.idxmin()
        seq = self._seq_holder.get_seq(
            df_alignment.loc[best_idx, 'query_name'],
            df_alignment.loc[best_idx, 'query_start'],
            df_alignment.loc[best_idx, 'query_end'],
            df_alignment.loc[best_idx, 'sstrand'],
        )
        if len(seq) == 0:
            # Allele coordinates extend beyond assembly boundaries —> likely contig edge
            return model.QueryResult(model.ALLELE_MISSING, [], tags=[model.Tag.EDGE])

        # Screen for imperfect matches
        best_matching = ImperfectMatchDetector(self._dir_db / locus_name)
        try:
            seq_ids_closest = best_matching.retrieve_best_matching(seq, self._min_id_novel)
        except InvalidLengthException:
            return model.QueryResult(model.ALLELE_MISSING, [], tags=[model.Tag.INDEL])

        # No imperfect hits to existing alleles found
        if len(seq_ids_closest) == 0:
            return model.QueryResult(model.ALLELE_MISSING, [], [])

        # Potential novel allele
        allele_hash = sequenceutils.hash_sequence(seq, rev_comp=False)
        allele_hash_shown = f'*{allele_hash[:4]}'
        return model.QueryResult(
            allele_str=allele_hash_shown,
            allele_results=[
                model.AlleleResult(
                    allele=allele_hash_shown,
                    alignment=model.Alignment(
                        seq_id=df_alignment.loc[best_idx, 'query_name'],
                        start=int(df_alignment.loc[best_idx, 'query_start']),
                        end=int(df_alignment.loc[best_idx, 'query_end']),
                        strand=df_alignment.loc[best_idx, 'sstrand'],
                    ),
                    length=len(seq),
                    sequence=seq,
                    closest_alleles=seq_ids_closest,
                )
            ],
            tags=[model.Tag.NOVEL],
        )

    def _process_locus(self, locus_name: str, df_alignment: pd.DataFrame) -> model.QueryResult:
        """
        Types the input locus.
        :param locus_name: Locus name
        :param df_alignment: Alignment data
        :return: Results for the locus
        """
        # Parse database information
        dir_locus = self._dir_db / locus_name
        with open(dir_locus / 'mist_db.json') as handle:
            data_locus = json.load(handle)

        # Process seed alignments
        matches = []
        for _, row in df_alignment.iterrows():
            # Retrieve the full sequence
            seq = self._seq_holder.get_seq(row['query_name'], row['query_start'], row['query_end'])
            if (seq is None) or (len(seq) == 0):
                continue

            # Check for an exact match
            match_perfect = self._extract_exact_match(seq, row, data_locus)
            if match_perfect is not None:
                matches.append(match_perfect)

        # Check if perfect matches have been found
        matches.sort(key=lambda res: data_locus['alleles'][res.allele]['idx'])
        if len(matches) > 0:
            return model.QueryResult(
                merge_results(matches, self._multi_strategy.value),
                allele_results=matches,
                tags=[model.Tag.MULTI] if len(matches) > 1 else [model.Tag.EXACT],
            )

        # Check imperfect matches
        logger.debug(f'Screening for imperfect hits for: {locus_name}')
        return self._extract_partial_match(df_alignment, locus_name)

    def query(
            self,
            path_fasta: Path,
            loci: list[str] | None = None,
            threads: int = 1) -> dict[str, model.QueryResult]:
        """
        Queries the database with the given FASTA file.
        :param path_fasta: Input FASTA file
        :param loci: List of target loci
        :param threads: Threads (number of threads to use)
        :return: Matching allele(s)
        """
        # Retrieve all loci
        with open(self._dir_db / 'loci.txt') as handle:
            all_loci = [l.strip() for l in handle]

        # Seed alignment
        logger.info('Performing seed alignment with Minimap2')
        data_mm2 = minimap2utils.align(
            path_fasta, self._dir_db / 'loci_repr.fasta', include_cigar=False, threads=threads)
        logger.info(f'{len(data_mm2):,} seed alignments')
        if self._save_minimap2:
            path_out = self._dir_out / 'minimap2_parsed.tsv'
            data_mm2.to_csv(path_out, sep='\t', index=False)
            logger.info(f'Saved minimap2 output to: {path_out}')

        # Check for empty results
        if len(data_mm2) == 0:
            logger.warning('No seed alignments found. Please verify that the correct scheme has been specified.')
            return {locus: model.QueryResult(model.ALLELE_MISSING, [], tags=[model.Tag.ABSENT]) for locus in all_loci}

        # Extract loci
        data_mm2['locus'] = data_mm2['sseqid'].str.rsplit('_', n=1).str[0]
        nb_loci = len(data_mm2['locus'].unique())
        logger.info(f"{nb_loci:,}/{len(all_loci):,} loci aligned ({100*nb_loci/len(all_loci):.2f}%)")

        # Calculate query string and remove duplicates
        data_mm2[['query_name', 'query_start', 'query_end']] = (
            data_mm2.apply(lambda x: AlleleQueryMinimap2.__get_coord(x), axis=1, result_type='expand'))
        data_mm2.drop_duplicates(['locus', 'query_name', 'query_start', 'query_end'], keep='first', inplace=True)
        logger.info(f'{len(data_mm2):,} seed alignments (without duplicates)')

        # Save the input sequence in memory
        self._seq_holder = SeqHolder(path_fasta)

        # Perform the querying
        results_by_locus: dict[str, model.QueryResult] = {}
        for locus, data in data_mm2.groupby('locus'):
            if (loci is not None) and (locus not in loci):
                continue
            results_by_locus[str(locus)] = self._process_locus(str(locus), data)
        return {locus: results_by_locus.get(locus, model.QueryResult(model.ALLELE_MISSING, [], [model.Tag.ABSENT])) for locus in all_loci}



import unittest
from importlib.resources import files
from pathlib import Path

from mist.app import model
from mist.app.query.allelequeryminimap import AlleleQueryMinimap2, MultiStrategy
from mist.app.utils import testingutils
from mist.scripts.mistindex import MistIndex


class TestAlleleQuery(unittest.TestCase):
    """
    Tests the allele calling method.
    """

    @staticmethod
    def get_expected_tag(allele_result: str | model.Tag) -> model.Tag:
        """
        Returns the expected tag for the given allele result.
        :param allele_result: Allele result
        :return: Expected tag
        """
        if isinstance(allele_result, model.Tag):
            return allele_result
        if allele_result == model.ALLELE_MISSING:
            return model.Tag.ABSENT
        if allele_result.startswith('*'):
            return model.Tag.NOVEL
        if '__' in allele_result:
            return model.Tag.MULTI
        return model.Tag.EXACT

    def setUp(self) -> None:
        """
        Sets up a temporary directory and builds a database there before each test.
        :return: None
        """
        paths_fasta = [
            Path(str(files('mist').joinpath('resources/testdata/NEIS0140-subset.fasta'))),
            Path(str(files('mist').joinpath('resources/testdata/NEIS0159-subset.fasta'))),
        ]
        self.dir_temp = testingutils.get_temp_dir()
        self.db_path = Path(self.dir_temp.name)

        # Build the index once for each test
        mist_idx = MistIndex(paths_fasta=paths_fasta, path_profiles=None)
        mist_idx.create_index(dir_out=self.db_path, threads=4)

    def tearDown(self) -> None:
        """
        Clean up the temporary directory after the test.
        :return: None
        """
        self.dir_temp.cleanup()

    def test_perfect_matches(self) -> None:
        """
        Tests allele calling with perfect matches for both loci.
        :return: None
        """
        caller = AlleleQueryMinimap2(dir_db=self.db_path)
        path_fasta = Path(str(files('mist').joinpath('resources/testdata/query-perfect_hits.fasta')))
        allele_by_locus = {'NEIS0140-subset': '10', 'NEIS0159-subset': '67'}
        result_by_locus = caller.query(path_fasta=path_fasta)
        for locus, res in result_by_locus.items():
            self.assertIn(model.Tag.EXACT, res.tags, f"No exact match for {locus}")
            self.assertEqual(res.allele_str, allele_by_locus[locus], f"Wrong allele for {locus}")

    def test_novel_allele(self) -> None:
        """
        Tests allele calling with a novel allele.
        :return: None
        """
        caller = AlleleQueryMinimap2(dir_db=self.db_path)
        path_fasta = Path(str(files('mist').joinpath('resources/testdata/query-novel.fasta')))
        allele_by_locus = {'NEIS0140-subset': '*dd0d', 'NEIS0159-subset': model.ALLELE_MISSING}
        result_by_locus = caller.query(path_fasta=path_fasta)
        for locus, res in result_by_locus.items():
            expected_tag = TestAlleleQuery.get_expected_tag(allele_by_locus[locus])
            self.assertIn(expected_tag, res.tags, f"No exact match for {locus}")
            self.assertEqual(res.allele_str, allele_by_locus[locus], f"Wrong allele for {locus}")

    def test_novel_allele_rc(self) -> None:
        """
        Tests allele calling with a novel allele on the reverse strand.
        :return: None
        """
        caller = AlleleQueryMinimap2(dir_db=self.db_path)
        path_fasta = Path(str(files('mist').joinpath('resources/testdata/query-novel_rc.fasta')))
        allele_by_locus = {'NEIS0140-subset': '*dd0d', 'NEIS0159-subset': model.ALLELE_MISSING}
        result_by_locus = caller.query(path_fasta=path_fasta)
        for locus, res in result_by_locus.items():
            expected_tag = TestAlleleQuery.get_expected_tag(allele_by_locus[locus])
            self.assertIn(expected_tag, res.tags, f"No exact match for {locus}")
            self.assertEqual(res.allele_str, allele_by_locus[locus], f"Wrong allele for {locus}")

    def test_multi_hit(self) -> None:
        """
        Tests allele calling with multiple exact hits.
        :return: None
        """
        caller = AlleleQueryMinimap2(dir_db=self.db_path, multi_strategy=MultiStrategy.ALL)
        path_fasta = Path(str(files('mist').joinpath('resources/testdata/query-multi_hit.fasta')))
        allele_by_locus = {'NEIS0140-subset': '1__108', 'NEIS0159-subset': model.ALLELE_MISSING}
        result_by_locus = caller.query(path_fasta=path_fasta)
        for locus, res in result_by_locus.items():
            expected_tag = TestAlleleQuery.get_expected_tag(allele_by_locus[locus])
            self.assertIn(expected_tag, res.tags, f"No exact match for {locus}")
            self.assertEqual(res.allele_str, allele_by_locus[locus], f"Wrong allele for {locus}")

    def test_all_absent(self) -> None:
        """
        Tests allele calling with no hits.
        :return: None
        """
        caller = AlleleQueryMinimap2(dir_db=self.db_path, multi_strategy=MultiStrategy.ALL)
        path_fasta = Path(str(files('mist').joinpath('resources/testdata/query-both_absent.fasta')))
        allele_by_locus = {'NEIS0140-subset': model.ALLELE_MISSING, 'NEIS0159-subset': model.ALLELE_MISSING}
        result_by_locus = caller.query(path_fasta=path_fasta)
        for locus, res in result_by_locus.items():
            expected_tag = TestAlleleQuery.get_expected_tag(allele_by_locus[locus])
            self.assertIn(expected_tag, res.tags, f"No exact match for {locus}")
            self.assertEqual(res.allele_str, allele_by_locus[locus], f"Wrong allele for {locus}")

    def test_one_absent(self) -> None:
        """
        Tests allele calling with no hits.
        :return: None
        """
        caller = AlleleQueryMinimap2(dir_db=self.db_path, multi_strategy=MultiStrategy.ALL)
        path_fasta = Path(str(files('mist').joinpath('resources/testdata/query-one_absent.fasta')))
        allele_by_locus = {'NEIS0140-subset': model.ALLELE_MISSING, 'NEIS0159-subset': '42'}
        result_by_locus = caller.query(path_fasta=path_fasta)
        for locus, res in result_by_locus.items():
            expected_tag = TestAlleleQuery.get_expected_tag(allele_by_locus[locus])
            self.assertIn(expected_tag, res.tags, f"No exact match for {locus}")
            self.assertEqual(res.allele_str, allele_by_locus[locus], f"Wrong allele for {locus}")

    def test_edge(self) -> None:
        """
        Tests allele calling with a partial hit on a contig edge.
        :return: None
        """
        caller = AlleleQueryMinimap2(dir_db=self.db_path, multi_strategy=MultiStrategy.ALL)
        path_fasta = Path(str(files('mist').joinpath('resources/testdata/query-edge.fasta')))
        allele_by_locus = {'NEIS0140-subset': model.Tag.EDGE, 'NEIS0159-subset': model.ALLELE_MISSING}
        result_by_locus = caller.query(path_fasta=path_fasta)
        for locus, res in result_by_locus.items():
            expected_tag = TestAlleleQuery.get_expected_tag(allele_by_locus[locus])
            self.assertIn(expected_tag, res.tags, f"No exact match for {locus}")

if __name__ == '__main__':
    unittest.main()



import unittest
from pathlib import Path

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from mist.app.loggers.logger import initialize_logging
from mist.app.utils import sequenceutils, testingutils


class TestSequenceUtils(unittest.TestCase):
    """
    Tests for the sequence utilities.
    """

    def test_hash_sequence(self) -> None:
        """
        Tests the hash sequence function.
        :return: None
        """
        hash_a = sequenceutils.hash_sequence('ACTGA', rev_comp=True)
        hash_b = sequenceutils.hash_sequence('CGATG', rev_comp=True)
        self.assertNotEqual(hash_a, hash_b)

    def test_rev_complement(self) -> None:
        """
        Tests the reverse complement function.
        :return: None
        """
        seq_in = 'ACTAGATAGAC'
        self.assertEqual(
            seq_in, sequenceutils.rev_complement(sequenceutils.rev_complement(seq_in))
        )

    def test_merge_fasta(self) -> None:
        """
        Tests the merge fasta function.
        :return: None
        """
        with testingutils.get_temp_dir() as dir_temp:
            # Create FASTA files
            path_fasta_a = Path(dir_temp, 'seqs_a.fasta')
            with path_fasta_a.open('w') as handle:
                SeqIO.write([SeqRecord(Seq('ACTG'), 'a1'), SeqRecord(Seq('CTGA'), 'a2')], handle, 'fasta')
            path_fasta_b = Path(dir_temp, 'seqs_b.fasta')
            with path_fasta_b.open('w') as handle:
                SeqIO.write([
                    SeqRecord(Seq('CCTG'), 'b1'),
                    SeqRecord(Seq('AAAA'), 'b2'),
                    SeqRecord(Seq('TTTT'), 'b3')],
                handle, 'fasta')

            # Merge FASTA files
            path_merged = Path(dir_temp, 'merged.fasta')
            nb_seqs = sequenceutils.merge_fasta_files([path_fasta_a, path_fasta_b], path_merged)
            self.assertEqual(nb_seqs, 5)


if __name__ == '__main__':
    initialize_logging()
    unittest.main()



#! /usr/bin/env python3
"""
Converts the MiST output to a partial LINcode.

The JSON input file should be generated with the MiST software and the Institut Pasteur Klebsiella scgMLST629_S scheme
(with cgMLST profiles!).

Example command:
mist call --fasta my_kleb_genome.fasta --db kleb_scgmlst_s --output my_kleb_genome.json
"""
import json
import sys

MISMATCH_CUTOFFS = [629, 610, 585, 190, 43, 10, 7, 4, 2, 1]
METADATA = [
    {'key': 'Phylogroup', 'bin_min': 2},
    {'key': 'Sublineage', 'bin_min': 3},
    {'key': 'Clonal group', 'bin_min': 4}
]


def determine_bin(nb_diff) -> int:
    """
    Determine the deepest bin index in the LINcode that can be reliably assigned.
    :param nb_diff: Number of differences
    :return: Bin number
    """
    for i, threshold in enumerate(MISMATCH_CUTOFFS):
        if i == len(MISMATCH_CUTOFFS) - 1:
            return i
        if threshold >= nb_diff >= MISMATCH_CUTOFFS[i+1]:
            return i
    raise ValueError('Cannot determine bin')

def create_lin_code(lin_code_in: str, bin_: int) -> str:
    """
    Construct the partial LINcode, inserting wildcard symbols for bins that cannot be reliably determined.
    :param lin_code_in: Input LINcode
    :param bin_: Bin number
    :return: Partial LINcode
    """
    lin_code_parts = lin_code_in.split('_')
    return '_'.join(lin_code_parts[:bin_] + (['*'] * (9 - bin_)))


if __name__ == '__main__':
    if len(sys.argv) != 2:
        raise ValueError('Usage: mist_to_partial_lincode.py <input_file.json>')
    with open(sys.argv[1]) as handle:
        data_in = json.load(handle)

    # Check if the input contains profile information
    if data_in['profile'] is None:
        raise ValueError('Profile information is missing')

    # Log the results
    best_matching_st = data_in['profile']['name']
    print(f'Best matching: scgST-{best_matching_st}')

    nb_loci = len(data_in['profile']['alleles'])
    try:
        nb_matches = data_in['profile']['nb_matches']
    except KeyError:
        nb_matches = int(data_in['profile']['pct_match'] * nb_loci / 100)
    print(f'Number of matches: {nb_matches}/{nb_loci}')

    # Parse the ST metadata
    metadata_st = {k: v for k, v in data_in['profile']['metadata']}
    lin_code = metadata_st['LINcode']
    print(f'LINcode for scgMST-{best_matching_st}: {lin_code}')

    # Construct the partial LINcode
    bin_nb = determine_bin(nb_loci - nb_matches)
    partial_lin_code = create_lin_code(lin_code, bin_nb)
    print(f'Partial LINcode for input strain: {partial_lin_code}')

    # Report the associated metadata
    for metadata_col in METADATA:
        value = data_in['profile']
        if metadata_col['bin_min'] > bin_nb:
            value = 'Cannot be determined'
        else:
            value = metadata_st[metadata_col['key']]
        print(f"{metadata_col['key']}: {value}")



import click
from furl import furl

from mist.app.dbs import BASE_URL_BY_DB, DOWNLOADERS


class MistList:
    """
    Commands to list the available schemes that can be downloaded.
    """

    def __init__(self, downloader: str, host: str, db: str | None = None) -> None:
        """
        Initializes the class.
        :param downloader: Key of the downloader class
        :param host: Host name
        :param db: Database name
        :return: None
        """
        download_key = downloader if downloader != 'bigsdb_auth' else 'bigsdb'
        self._downloader = DOWNLOADERS[download_key]()
        self._downloader_key = downloader
        self._host = host if host is not None else 'base'
        self._db = db

    def _build_bigsdb_url(self) -> str:
        """
        Returns the BIGSdb URL.
        :return: BIGSdb URL
        """
        if self._db is None:
            return BASE_URL_BY_DB['bigsdb'][self._host]
        else:
            api_root = BASE_URL_BY_DB['bigsdb'][self._host]
            return str(furl(api_root).add(path='db').add(path=self._db).add(path='schemes'))

    def print_available_schemes(self) -> None:
        """
        Prints the available schemes and the corresponding URL.
        :return: None
        """
        # Determine the base URL
        if self._downloader_key in ('bigsdb', 'bigsdb_auth'):
            base_url = furl(self._build_bigsdb_url())
        else:
            base_url = furl(BASE_URL_BY_DB[self._downloader_key][self._host])

        # Retrieve the available schemes
        schemes = self._downloader.get_available_schemes(furl(base_url), species=self._db is None)
        if len(schemes) ==0:
            click.echo('No schemes found')
            raise RuntimeError('No schemes found')

        # Log the results
        for scheme_name, url in schemes:
            click.echo(f"{scheme_name:<30} {url}")
        click.echo('-' * 10)

        if (self._db is None) and ('bigsdb' in self._downloader_key):
            click.echo(
                "IMPORTANT: Due to the large number of schemes hosted on PubMLST, they are grouped by species.\n"
                "To list the available databases for a species, use:\n"
                f"mist list --downloader {self._downloader_key} --host {self._host} --db {{DB}}\n"
                "where {DB} is the name of the database from the list above (with the '_seqdef' suffix)"
            )
        else:
            click.echo(
                'Download the scheme using:\n'
                f'mist download --downloader {self._downloader_key} --url {{URL}} --output {{DIR}}')



from io import StringIO
from pathlib import Path
from typing import TextIO

import pandas as pd

from mist.app.utils.command import Command

COLS_PAF = {
    'qseqid': str,
    'qlen': int,
    'qstart': int,
    'qend': int,
    'sstrand': str,
    'sseqid': str,
    'slen': int,
    'sstart': int,
    'send': int,
    'bases_matching': int,
    'bases_total': int,
    'mq': int,
}


def create_index(path_fasta: Path) -> Path:
    """
    Creates a Minimap2 index for the input FASTA file.
    :param path_fasta: Input FASTA file
    :return: None
    """
    path_out = path_fasta.parent / f'{path_fasta.name}.mni'
    command = Command(' '.join([
        'minimap2',
        '-d', str(path_out),
        str(path_fasta)
    ]))
    command.run(path_fasta.parent, disable_logging=False)
    if not command.exit_code == 0:
        raise RuntimeError(f'Error creating Minimap2 index: {command.stderr}')
    return path_out

def align(path_fasta: Path, path_fasta_db: Path, include_cigar: bool = False, threads: int = 1) -> pd.DataFrame:
    """
    Runs Minimap2 on the input sequence and DB.
    :param path_fasta: Input FASTA path
    :param path_fasta_db: Database FASTA path
    :param include_cigar: Include cigar string in output
    :param threads: Number of threads to use
    :return: Output results as DataFrame
    """
    command = Command(' '.join([
        'minimap2',
        str(path_fasta_db),
        str(path_fasta),
        '-O 25 -E 1', # ungapped alignment
        '-N 10 --all-chain', # include secondary alignments
        '-t', str(threads) # nb. of threads
    ] + (['-c'] if include_cigar else [])))
    command.run(Path().cwd(), disable_logging=False)
    if not command.exit_code == 0:
        raise RuntimeError(f'Error running Minimap2: {command.stderr}')
    return parse_paf(StringIO(command.stdout))


def parse_paf(file_in: Path | TextIO) -> pd.DataFrame:
    """
    Parses the input PAF file.
    :param file_in: Input path (or file-like object)
    :return: Parsed data as DataFrame
    """
    records_out = []
    offset = len(COLS_PAF)

    handle = file_in.open() if isinstance(file_in, Path) else file_in
    with handle:
        for row in handle:
            parts = row.rstrip('\n').split('\t')
            record = {col: type_(parts[i]) for i, (col, type_) in enumerate(COLS_PAF.items())}
            for tag in parts[offset:]:
                tag_name = tag.split(':', 1)[0]
                record[f"tag_{tag_name}"] = tag
            records_out.append(record)
    return pd.DataFrame(records_out)



import re
from pathlib import Path

from Bio import SeqIO

from mist.app.loggers.logger import logger


def is_valid_db(dir_in: Path) -> bool:
    """
    Checks if the input directory is a valid typing database.
    :param dir_in: Input directory
    :return: True if valid
    """
    if not (dir_in / 'loci.txt').exists():
        raise FileNotFoundError("'loci.txt' file not found")
    if not (dir_in / 'loci_repr.fasta').exists():
        raise FileNotFoundError("'loci_repr.fasta' file not found")
    return True

def _get_allele_id(seq_record: SeqIO.SeqRecord, locus_name: str, requires_match: bool = False) -> str:
    """
    Returns the allele id from the input SeqRecord
    :param seq_record: Sequence record
    :param locus_name: Locus name
    :param requires_match: If True, the sequence id must match the locus name
    :return: Allele id
    """
    if re.match(r'\d+', seq_record.id):
        return f'{locus_name}_{seq_record.id}'
    if re.match(f'{locus_name}_\\d+', seq_record.id, flags=re.IGNORECASE):
        return seq_record.id
    m = re.match(r'.+_(\d+)$', seq_record.id)
    if not requires_match and (m is not None):
        return f'{locus_name}_{m.group(1)}'
    raise ValueError(f'Invalid sequencing input FASTA file: {seq_record.id}')


def reformat_fasta(fasta_in: Path, fasta_out: Path) -> None:
    """
    Reforms the input FASTA file to the standardized format required by the tool.
    :param fasta_in: Input FASTA file
    :param fasta_out: Output FASTA file
    :return: None
    """
    if not fasta_in.name.endswith('.fasta'):
        raise ValueError("Input FASTA files should have the '.fasta' extension")
    locus_name = fasta_in.name.replace('.fasta', '')
    seq_ids_match_name = True
    with open(fasta_in) as handle_in, open(fasta_out, 'w') as handle_out:
        for seq in SeqIO.parse(handle_in, 'fasta'):
            try:
                allele_id = _get_allele_id(seq, locus_name, requires_match=True)
            except ValueError:
                seq_ids_match_name = False
                allele_id = _get_allele_id(seq, locus_name, requires_match=False)
            seq_out = SeqIO.SeqRecord(
                id=allele_id,
                description='',
                seq=seq.seq
            )
            SeqIO.write(seq_out, handle_out, 'fasta')
    if not seq_ids_match_name:
        logger.warning(
            f"FASTA file '{fasta_in.name}' contains sequences that do not match the locus name '{locus_name}'.")



import re
from pathlib import Path
from typing import TypedDict

from mist.app.loggers.logger import logger
from mist.app.utils.command import Command


class ClusterDict(TypedDict):
    """
    Holds cluster information.
    """
    name: str
    id: int
    members: list[dict[str, str]]


def cluster_fasta(path_fasta: Path, dir_out: Path, cutoff: int, threads: int = 4, debug: bool = False) -> Path:
    """
    Clusters the input FASTA file.
    :param path_fasta: Input FASTA file
    :param dir_out: Output directory
    :param cutoff: Clustering cutoff
    :param threads: Number of threads
    :param debug: If True, the command is logged
    :return: Path to the clustered FASTA file
    """
    path_out = dir_out / path_fasta.name.replace(".fasta", "-clustered.fasta")
    command = Command(
        ' '.join([
            'cd-hit-est',
            '-i', str(path_fasta),
            '-o', str(path_out),
            '-S 0',
            '-d 0',
            '-c', str(cutoff / 100),
            '-T', str(threads)])
    )
    command.run(dir_out, disable_logging=not debug)
    if not command.exit_code == 0:
        raise ValueError(f"error running clustering: {command.stderr}")
    return path_out


def parse_cluster_from_file(path_fasta: Path) -> list[ClusterDict]:
    """
    Parses a CD-HIT .clstr file and returns a list of Cluster objects.
    :param path_fasta: Path to the CD-hit clustered FASTA file
    :return: List of clusters
    """
    # Check for the file with clusters
    path_clusters = path_fasta.parent / f'{path_fasta.name}.clstr'
    if not path_clusters.exists():
        raise FileNotFoundError(f'Clusters file not found: {path_clusters}')

    clusters = []
    re_cluster = re.compile(r'\d.+, >(.+)\.\.\. (at [+-])?.*$')

    current_cluster = None
    with path_clusters.open() as handle:
        for current_line in handle.readlines():
            line = current_line.strip()
            if line.startswith('>'):
                try:
                    name = line[1:].replace(' ', '_')
                    cluster_id = int(name.split('_')[-1])
                except (ValueError, IndexError):
                    raise ValueError(f'Invalid cluster header: {line}')
                current_cluster = {'name': name, 'id': cluster_id, 'members': []}
                clusters.append(current_cluster)
            else:
                if current_cluster is None:
                    raise ValueError(f'Sequence line found before any cluster header: {line}')
                match = re_cluster.match(line)
                if not match:
                    raise ValueError(f'Invalid cluster line: {line}')
                current_cluster['members'].append({
                    'seq_id': match.group(1), 'ori': match.group(2)[-1] if match.group(2) else None})
    logger.debug(f'{len(clusters):,} clusters parsed from {path_fasta.name}')
    return clusters



from pathlib import Path
from typing import Any

from furl import furl

from mist.app.dbs.basedownloader import BaseDownloader
from mist.app.loggers.logger import logger
from mist.app.utils import restutils, sequenceutils


class BIGSDbDownloader(BaseDownloader):
    """
    Downloader for BIGSdb databases.
    Note: this downloader does **NOT** handle authorization and will not download all available data.
    """

    DOWNLOADER_KEY = 'bigsdb'

    def _download_fasta_files(self, metadata: dict[str, Any]) -> list[Path]:
        """
        Downloads the scheme FASTA files.
        :param metadata: Scheme metadata
        :return: List of FASTA paths
        """
        paths_fasta = []
        for href_locus in metadata['loci']:
            locus = furl(href_locus).path.segments[-1]
            url_fasta = str(furl(href_locus).add(path='alleles_fasta'))
            path_fasta = restutils.download_fasta(locus_name=locus, url=url_fasta, dir_out=self.dir_out)
            paths_fasta.append(path_fasta)
            logger.info(
                f'Downloaded {path_fasta.name} ({sequenceutils.count_sequences(path_fasta):,} sequences)'
            )
        return paths_fasta

    def _download(self, url: str, dir_out: Path, include_profiles: bool = False) -> None:
        """
        Downloads the target scheme.
        :param url: Scheme URL
        :param dir_out: Output directory
        :param include_profiles: Include profiles
        :return: None
        """
        logger.warning('Downloading from PubMLST **without** authentication, not all data will be retrieved!')

        # Download the scheme metadata
        metadata = restutils.retrieve_page_data(url).json()
        logger.info(f"Metadata received: {metadata['locus_count']:,} loci")

        # Download FASTA files
        paths_fasta = self._download_fasta_files(metadata)

        # Download profiles
        if include_profiles:
            href_profiles = str(furl(url).add(path='profiles_csv'))
            response = restutils.retrieve_page_data(href_profiles)
            with open(self.dir_out / 'profiles.tsv', 'w') as handle:
                handle.write(response.text)
            logger.info('Profiles downloaded')

        # Create a TXT file with all FASTA files
        self.create_fasta_list(paths_fasta)
        logger.info('Scheme downloaded successfully')

    def get_available_schemes(self, base_url: furl, **kwargs: Any) -> list[tuple[str, str]]:
        """
        Retrieve a list of available cgMLST schemes and their URL.
        :param base_url: Base URL
        :param kwargs: Keyword arguments
        :return: List of schemes / species and URL
        """
        data = restutils.retrieve_page_data(str(base_url))

        # Species
        schemes = []

        if kwargs.get('species', False) is True:
            for row in data.json():
                # Retrieve the available species
                try:
                    db_info = next(d for d in row['databases'] if d['name'].endswith('_seqdef'))
                except StopIteration:
                    logger.debug(f"No SeqDef database found for: {row['description']}")
                    continue
                schemes.append((row['description'].strip(), db_info['href'].split('/')[-1]))
        else:
            for data_scheme in data.json()['schemes']:
                schemes.append((data_scheme['description'], data_scheme['scheme']))
        return schemes



import unittest
from importlib.resources import files
from pathlib import Path

from mist.app.loggers.logger import initialize_logging
from mist.app.utils import clusterutils, testingutils


class TestClusteringUtils(unittest.TestCase):
    """
    Tests for the clustering utilities.
    """

    def test_clustering_with_cdhit(self) -> None:
        """
        Tests the CD-HIT clustering function.
        :return: None
        """
        with testingutils.get_temp_dir() as dir_temp:
            path_in = Path(str(files('mist').joinpath('resources/testdata/NEIS0140-subset.fasta')))
            clusterutils.cluster_fasta(path_in, Path(dir_temp), 90)

    def test_clustering_with_cdhit_parsing(self) -> None:
        """
        Tests the CD-HIT clustering function with parsing.
        :return: None
        """
        with testingutils.get_temp_dir() as dir_temp:
            path_in = Path(str(files('mist').joinpath('resources/testdata/NEIS0140-subset.fasta')))
            path_fasta = clusterutils.cluster_fasta(path_in, Path(dir_temp), 90)
            clusters = clusterutils.parse_cluster_from_file(path_fasta)
            self.assertGreater(len(clusters), 0)
            self.assertEqual(sum(len(c['members']) for c in clusters), 5)


if __name__ == '__main__':
    initialize_logging()
    unittest.main()



import unittest
from importlib.resources import files
from pathlib import Path

from click.testing import CliRunner

from mist.app.loggers.logger import initialize_logging, logger
from mist.app.utils import dbutils, testingutils
from mist.scripts.cli import cli


class TestIndex(unittest.TestCase):
    """
    Tests for the indexing functionality.
    """

    def test_index_single_fasta(self) -> None:
        """
        Tests indexing on a single fasta file.
        :return: None
        """
        runner = CliRunner()
        with testingutils.get_temp_dir() as dir_temp:
            # noinspection PyTypeChecker
            result = runner.invoke(
                cli,[
                    'index',
                    str(files('mist').joinpath('resources/testdata/NEIS0140-subset.fasta')),
                    '--output', str(dir_temp),
                    '--threads', '4'
                ], catch_exceptions=False
            )
            logger.info(result.output)
            self.assertTrue(result.exit_code == 0)
            self.assertTrue(dbutils.is_valid_db(Path(dir_temp)))

    def test_index_single_fasta_debug(self) -> None:
        """
        Tests indexing on a single fasta file in debug mode.
        :return: None
        """
        runner = CliRunner()
        with testingutils.get_temp_dir() as dir_temp:
            # noinspection PyTypeChecker
            result = runner.invoke(
                cli, [
                    'index',
                    str(files('mist').joinpath('resources/testdata/NEIS0140-subset.fasta')),
                    '--output', str(dir_temp),
                    '--threads', '4',
                    '--debug',
                ], catch_exceptions=False
            )
            logger.info(result.output)
            self.assertTrue(dbutils.is_valid_db(Path(dir_temp)))
            self.assertEqual(result.exit_code, 0)

    def test_index_single_fasta_diff_fmt(self) -> None:
        """
        Tests indexing on a single fasta file in a different format.
        Each allele has just an identifier instead of the full allele name:
        >1
        [SEQ]
        >2
        [SEQ]
        ...
        :return: None
        """
        runner = CliRunner()
        with testingutils.get_temp_dir() as dir_temp:
            # noinspection PyTypeChecker
            result = runner.invoke(
                cli, [
                    'index',
                    str(files('mist').joinpath('resources/testdata/NEIS0140-fmt.fasta')),
                    '--output', str(dir_temp),
                    '--threads', '4'
                ], catch_exceptions=False
            )
            logger.info(result.output)
            self.assertTrue(dbutils.is_valid_db(Path(dir_temp)))
            self.assertEqual(result.exit_code, 0)

    def test_index_from_list(self) -> None:
        """
        Tests indexing FASTA files from a list.
        :return: None
        """
        runner = CliRunner()
        with testingutils.get_temp_dir() as dir_temp:
            # Create a TXT file with the FASTA files
            path_txt = Path(dir_temp, 'fasta_in.txt')
            with path_txt.open('w') as handle:
                handle.write(str(files('mist').joinpath('resources/testdata/NEIS0140-fmt.fasta')) + '\n')
                handle.write(str(files('mist').joinpath('resources/testdata/NEIS0140-subset.fasta')) + '\n')

            # Index the database
            # noinspection PyTypeChecker
            result = runner.invoke(
                cli, [
                    'index',
                    '--fasta-list', str(path_txt),
                    '--output', str(dir_temp),
                    '--threads', '4',
                ], catch_exceptions=False
            )
            logger.info(result.output)
            self.assertTrue(dbutils.is_valid_db(Path(dir_temp)))
            self.assertEqual(result.exit_code, 0)


if __name__ == '__main__':
    initialize_logging()
    unittest.main()



import time
import zipfile
from pathlib import Path
from typing import Any

import bs4
import requests
from furl import furl

from mist.app.dbs.basedownloader import BaseDownloader
from mist.app.loggers.logger import logger
from mist.app.utils.restutils import retrieve_page_data


class CgMLSTOrgDownloader(BaseDownloader):
    """
    Downloader for cgMLST.org databases.
    """

    DOWNLOADER_KEY = 'cgmlstorg'

    @staticmethod
    def retrieve_page_data(
        url: str, retries: int = 3, timeout: int = 60 * 5
    ) -> requests.Response:
        """
        Retrieves data from the given URL.
        :param url: URL
        :param retries: Number of retries
        :param timeout: Timeout (in seconds)
        :return: URL data
        """
        error = None
        for retry in range(retries):
            try:
                response = requests.get(url, timeout=timeout)
                response.raise_for_status()
                return response
            except BaseException as err:
                logger.warning(
                    f'Error retrieving page data ({err}), retrying (attempt {retry + 1})'
                )
                error = err
                time.sleep(1)
        raise RuntimeError(f'Error retrieving page data: {url}: {error}')

    @staticmethod
    def retrieve_fasta_archive(url: str, output_dir: Path) -> Path:
        """
        Retrieves the FASTA archive from the cgmlst.org website
        :param url: URL
        :param output_dir: Output directory
        :return: Path to the output FASTA
        """
        output_path = output_dir / 'fasta.zip'
        with open(output_path, 'wb') as handle:
            handle.write(CgMLSTOrgDownloader.retrieve_page_data(url).content)
        logger.info(f'Archive downloaded to: {output_path}')
        return output_path

    @staticmethod
    def extract_fasta_files(input_archive: Path, output_dir: Path) -> list[Path]:
        """
        Extracts the FASTA files from the zip archive
        :param input_archive: Downloaded zip archive
        :param output_dir: Output directory
        :return: List of paths to the FASTA files
        """
        with zipfile.ZipFile(input_archive, 'r') as handle:
            if handle.testzip() is not None:
                raise zipfile.BadZipFile(f"Corrupt zip file: {input_archive}")
            handle.extractall(output_dir)
        return [f for f in output_dir.iterdir() if f.name.endswith('.fasta')]

    def _download(self, url: str, dir_out: Path, include_profiles: bool = False, **kwargs: Any) -> None:
        """
        Downloads the target scheme.
        :param url: Scheme URL
        :param dir_out: Output directory
        :param include_profiles: Include profiles
        :param kwargs: Additional arguments
        :return: None
        """
        # Download the archive containing the FASTA files
        href_alleles = str(furl(url).add(path='alleles'))
        logger.info('Starting download of FASTA files')
        path_fasta_arch = CgMLSTOrgDownloader.retrieve_fasta_archive(href_alleles, dir_out)
        paths_fasta = CgMLSTOrgDownloader.extract_fasta_files(path_fasta_arch, dir_out)
        logger.info(f"Found {len(paths_fasta):,} FASTA files")

        # Download the profiles (if specified)
        if include_profiles:
            href_profiles = str(furl(url).add(path='locus/?content-type=csv'))
            with open(self.dir_out / 'profiles.tsv', 'w') as handle:
                handle.write(retrieve_page_data(href_profiles).text)
            logger.info('Profiles downloaded')

        # Create a TXT file with all FASTA files
        self.create_fasta_list(paths_fasta)
        logger.info('Scheme downloaded successfully')

    def get_available_schemes(self, base_url: furl, **kwargs: Any) -> list[tuple[str, str]]:
        """
        Retrieve a list of available cgMLST schemes and their URL.
        :param base_url: Base URL
        :param kwargs: Keyword arguments
        :return: None
        """
        response = retrieve_page_data(str(base_url))
        soup = bs4.BeautifulSoup(response.text, 'html.parser')
        rows = soup.select('table.table.table-striped tbody tr')
        schemes = []
        for row in rows:
            cells = row.select('td')
            url = cells[0].a['href']
            scheme_name = cells[0].get_text(strip=True)
            schemes.append((scheme_name, url))
        return schemes



import json
import unittest
from importlib.resources import files
from pathlib import Path

from click.testing import CliRunner

from mist.app.loggers.logger import initialize_logging
from mist.app.utils import sequenceutils, testingutils
from mist.scripts.cli import cli
from mist.scripts.mistindex import MistIndex


class TestCall(unittest.TestCase):
    """
    Tests the allele calling script.
    """

    def setUp(self) -> None:
        """
        Sets up a temporary directory and builds a database there before each test.
        :return: None
        """
        paths_fasta = [
            Path(str(files('mist').joinpath('resources/testdata/NEIS0140-subset.fasta'))),
            Path(str(files('mist').joinpath('resources/testdata/NEIS0159-subset.fasta'))),
        ]
        self.dir_temp = testingutils.get_temp_dir()
        self.db_path = Path(self.dir_temp.name)

        # Build the index once for each test
        mist_idx = MistIndex(paths_fasta=paths_fasta, path_profiles=None)
        mist_idx.create_index(dir_out=self.db_path, threads=4)

    def tearDown(self) -> None:
        """
        Clean up the temporary directory after the test.
        :return: None
        """
        self.dir_temp.cleanup()

    def test_call_with_hits(self) -> None:
        """
        Tests calling the alleles with all perfect hits.
        :return: None
        """
        runner = CliRunner()
        with testingutils.get_temp_dir() as dir_temp:
            # Output file(s)
            dir_out = Path(dir_temp, 'out')
            dir_out.mkdir(parents=True, exist_ok=True)
            path_json = dir_out / 'alleles.json'

            # Run the allele calling
            # noinspection PyTypeChecker
            result = runner.invoke(
                cli,
                [
                    'call',
                    '--fasta', str(files('mist').joinpath('resources/testdata/query-perfect_hits.fasta')),
                    '--db', str(self.db_path),
                    '--out-json', str(path_json),
                    '--threads', '4'
                ], catch_exceptions=False
            )

            # Check if the command was executed successfully
            self.assertEqual(0, result.exit_code)
            self.assertTrue(path_json.exists(), "Output JSON file not generated")

    def test_call_with_novel_allele(self) -> None:
        """
        Tests querying the database with a novel hit.
        :return: None
        """
        runner = CliRunner()
        with testingutils.get_temp_dir() as dir_temp:
            # Output file(s)
            dir_out = Path(dir_temp, 'out')
            dir_out.mkdir(parents=True, exist_ok=True)
            path_json = dir_out / 'alleles.json'

            # Run the script
            # noinspection PyTypeChecker
            result = runner.invoke(
                cli, [
                    'call',
                    '--fasta', str(files('mist').joinpath('resources/testdata/query-novel.fasta')),
                    '--db', str(self.db_path),
                    '--out-json', str(path_json),
                    '--out-dir', str(dir_out),
                    '--threads', '4',
                    '--export-novel'
            ], catch_exceptions=False)

            # Check for the FASTA output
            self.assertTrue((dir_out / 'novel_alleles').exists(), "novel_alleles directory not found")
            path_fasta = next((dir_out / 'novel_alleles').glob('*.fasta'))
            self.assertEqual(sequenceutils.count_sequences(path_fasta), 1)
            self.assertEqual(result.exit_code, 0)

    def test_call_with_novel_allele_rc(self) -> None:
        """
        Tests querying the database with a novel hit.
        Ensures that the same allele id is obtained regardless of strand
        :return: None
        """
        runner = CliRunner()
        fasta_in = [
            str(files('mist').joinpath('resources/testdata/query-novel.fasta')),
            str(files('mist').joinpath('resources/testdata/query-novel_rc.fasta'))
        ]
        calls_out = []
        with testingutils.get_temp_dir() as dir_temp:
            # Run the calling
            for i, path_fasta in enumerate(fasta_in):
                # Output file(s)
                dir_out = Path(dir_temp, f'out_{i}')
                dir_out.mkdir(parents=True, exist_ok=True)
                path_json = dir_out / 'alleles.json'

                # Run the script
                # noinspection PyTypeChecker
                result = runner.invoke(
                    cli, [
                        'call',
                        '--fasta', str(path_fasta),
                        '--db', str(self.db_path),
                        '--out-json', str(path_json),
                        '--threads', '4',
                    ],
                )
                self.assertEqual(result.exit_code, 0)
                with open(path_json) as handle:
                    calls_out.append(json.load(handle))

        self.assertEqual(
            calls_out[0]['alleles']['NEIS0140-subset']['allele_str'],
            calls_out[1]['alleles']['NEIS0140-subset']['allele_str'],
            "Allele hashes do not match"
        )

    def test_call_no_hit(self) -> None:
        """
        Tests querying the database without a hit.
        :return: None
        """
        runner = CliRunner()
        with testingutils.get_temp_dir() as dir_temp:
            # Output file(s)
            dir_out = Path(dir_temp, 'out')
            dir_out.mkdir(parents=True, exist_ok=True)
            path_json = dir_out / 'alleles.json'

            # Run the script
            # noinspection PyTypeChecker
            result = runner.invoke(
                cli,[
                    'call',
                    '--fasta', str(files('mist').joinpath('resources/testdata/query-both_absent.fasta')),
                    '--db', str(self.db_path),
                    '--out-json', str(path_json),
                    '--threads', '4',
                ], catch_exceptions=False
            )
            self.assertEqual(result.exit_code, 0)


if __name__ == '__main__':
    initialize_logging()
    unittest.main()



import itertools
import json
from pathlib import Path

import pandas as pd

from mist.app.loggers.logger import logger


class MistDists:
    """
    Class for generating distance and allele matrices from typing outputs.
    Expected input:
    - a list of TSV or JSON files generated by MiST
    """

    MIN_NB_DATASETS = 3

    def __init__(
            self, inputs: list[Path], out_matrix: Path, out_dists: Path, min_perc_loci: int,
            min_perc_samples: int) -> None:
        """
        Initializes the main script.
        :param inputs: List of input files
        :param out_matrix: Output allele matrix
        :param out_dists: Output file with pairwise distances
        :param min_perc_loci: Remove datasets with <% of loci detected
        :param min_perc_samples: Remove loci detected in <% of samples
        :return: None
        """
        self._inputs = inputs
        self._out_matrix = out_matrix
        self._out_dists = out_dists
        self._min_perc_loci = min_perc_loci
        self._min_perc_samples = min_perc_samples

    @staticmethod
    def _calc_distance(alleles_a: pd.Series, alleles_b: pd.Series) -> int:
        """
        Calculate the pairwise allele distance.
        :param alleles_a: Alleles a
        :param alleles_b: Alleles b
        :return: Number of allelic differences
        """
        total_dist = 0
        for allele_a, allele_b in zip(alleles_a, alleles_b):
            if allele_a == '-' and allele_b == '-':
                continue
            else:
                total_dist += 0 if allele_a == allele_b else 1
        return total_dist

    @staticmethod
    def parse_tsv(path_in: Path) -> tuple[str, pd.Series]:
        """
        Parses the input TSV file.
        :param path_in: Input path
        :return: Parsed data
        """
        logger.debug(f'Parsing file: {path_in}')
        data_tsv = pd.read_table(path_in, dtype=str, index_col='locus')
        return path_in.name.replace('.tsv', ''), data_tsv['allele']

    @staticmethod
    def parse_json(path_in: Path) -> tuple[str, pd.Series]:
        """
        Parses the input JSON file.
        :param path_in: Input path
        :return: Parsed data
        """
        logger.debug(f'Parsing file: {path_in}')
        with path_in.open() as handle:
            data = json.load(handle)
        alleles = pd.Series({locus: row['allele_str'] for locus, row in data['alleles'].items()})
        return path_in.name.replace('.json', ''), alleles

    def _log_nb_perfect_hits(self, name, alleles: pd.Series) -> None:
        """
        Logs the number and percentage of perfect hits for the input dataset.
        :param name: Dataset name
        :param alleles: Allele calls
        :return: None
        """
        nb_perfect = len(alleles) - alleles.eq('-').sum()
        perc_perfect = 100 * nb_perfect / len(alleles)
        logger.info(f"{name}: {nb_perfect}/{len(alleles):,} ({perc_perfect:.2f}%) perfect hits")

    def _parse_input_files(self) -> pd.DataFrame:
        """
        Parse the input files.
        :return: DataFrame with detected alleles (index = sample name)
        """
        # Parse the input
        datasets = {}
        for path_in in self._inputs:
            if path_in.suffix == '.tsv':
                name, alleles = MistDists.parse_tsv(path_in)
                datasets[name] = alleles
            elif path_in.suffix == '.json':
                name, alleles = MistDists.parse_json(path_in)
                datasets[name] = alleles
            else:
                raise ValueError(f'Invalid extension: {path_in.suffix} (expected .tsv or .json)')
            self._log_nb_perfect_hits(name, alleles)
        if len(datasets) < MistDists.MIN_NB_DATASETS:
            raise ValueError(f'At least {MistDists.MIN_NB_DATASETS} input files are required (found: {len(datasets)})')

        # Create merged DataFrame
        return pd.DataFrame(data=list(datasets.values()), index=list(datasets.keys()), dtype=str)

    def _calc_distance_matrix(self, allele_data_filt: pd.DataFrame) -> pd.DataFrame:
        """
        Calculates the pairwise distance matrix
        :param allele_data_filt: Filtered allele matrix
        :return: Distance matrix
        """
        # Calculate pair-wise distances
        distance_by_dataset_pair = {}
        for dataset_a, dataset_b in itertools.combinations(allele_data_filt.index, r=2):
            key = tuple(sorted([dataset_a, dataset_b]))
            dist = MistDists._calc_distance(allele_data_filt.loc[dataset_a], allele_data_filt.loc[dataset_b])
            distance_by_dataset_pair[key] = dist

        # Create data frame with pairwise distances
        records_out = []
        for dataset_a in allele_data_filt.index:
            records_out.append({
                dataset_b: distance_by_dataset_pair.get(tuple(sorted([dataset_a, dataset_b])), 0) for
                dataset_b in allele_data_filt.index
            })
        return pd.DataFrame(records_out, index=allele_data_filt.index)

    def _filt_allele_matrix(self, allele_data: pd.DataFrame) -> pd.DataFrame:
        """
        Filters the allele matrix by removing:
        - Datasets with less than x% of loci detected
        - Loci present in less than x% of datasets
        :param allele_data: Allele data
        :return: Filtered allele data, loci cutoff, datasets cutoff
        """
        # Filter allele matrix (nb. of loci detected per dataset)
        nb_loci_detected = allele_data.apply(lambda x: len(x) - list(x).count('-'), axis=1)
        cutoff_loci = int(self._min_perc_loci * len(allele_data.columns) / 100)
        logger.info(f"Removing datasets with <{cutoff_loci} ({self._min_perc_loci}%) loci detected")
        allele_data_filt = allele_data[nb_loci_detected > cutoff_loci]
        logger.info(f"{len(allele_data_filt)}/{len(allele_data)} datasets passed filtering")

        # Filter allele matrix (loci detected in nb. of datasets)
        cutoff_datasets = int(self._min_perc_samples * len(allele_data_filt) / 100)
        logger.info(f"Removing loci detected in <{cutoff_datasets} ({self._min_perc_samples}%) samples")
        # noinspection PyUnresolvedReferences
        locus_present_in_datasets = (allele_data_filt != '-').sum(axis=0)
        allele_data_filt = allele_data_filt.loc[:, locus_present_in_datasets > cutoff_datasets]
        logger.info(f"{len(allele_data_filt.columns)}/{len(allele_data.columns)} loci passed filtering")
        return allele_data_filt

    def run(self) -> None:
        """
        The main method to construct the MLST phylogeny.
        :return: None
        """
        # Parse the input files
        df_alleles = self._parse_input_files()

        # Filter the allele matrix
        df_alleles_filt = self._filt_allele_matrix(df_alleles)
        df_alleles_filt.to_csv(self._out_matrix, index_label='ID', sep='\t')
        logger.info(f'Allele matrix exported to: {self._out_matrix}')

        # Calculate the distance matrix
        df_dists = self._calc_distance_matrix(df_alleles_filt)

        # Check the distance matrix
        if df_dists.to_numpy().max() == 0:
            raise ValueError('Empty distance matrix')

        # Export the distance matrix
        df_dists.to_csv(self._out_dists, index_label='ID', sep='\t')
        logger.info(f'Distance matrix exported to: {self._out_dists}')

        # Log the command for GrapeTree
        logger.info(f'You can construct a phylogeny using: grapetree --profile {self._out_matrix} --method MSTreeV2')



import unittest
from importlib.resources import files
from pathlib import Path

from mist.app.loggers.logger import initialize_logging
from mist.app.utils import minimap2utils, testingutils


class TestMinimap2Utils(unittest.TestCase):
    """
    Tests for the Minimap2 utilities.
    """

    def test_minimap2_index(self) -> None:
        """
        Tests the create_index function.
        :return: None
        """
        path_db = Path(str(files('mist').joinpath('resources/testdata/NEIS0140-subset.fasta')))
        with testingutils.get_temp_dir() as dir_temp:
            path_in = Path(dir_temp, path_db.name)
            path_in.symlink_to(path_db)
            path_index = minimap2utils.create_index(path_in)
            self.assertGreater(path_index.stat().st_size, 0)

    def test_minimap2_index_and_query(self) -> None:
        """
        Tests the create_index function followed by alignment.
        :return: None
        """
        path_db = Path(str(files('mist').joinpath('resources/testdata/NEIS0140-subset.fasta')))
        path_query = Path(str(files('mist').joinpath('resources/testdata/query-perfect_hits.fasta')))
        with testingutils.get_temp_dir() as dir_temp:
            # Create index
            path_in = Path(dir_temp, path_db.name)
            path_in.symlink_to(path_db)
            path_index = minimap2utils.create_index(path_in)
            self.assertGreater(path_index.stat().st_size, 0)

            # Align to index
            data_out = minimap2utils.align(path_query, path_in)
            self.assertGreater(len(data_out), 0)

    def test_minimap2_query(self) -> None:
        """
        Tests minimap2 query function.
        :return: None
        """
        path_db = Path(str(files('mist').joinpath('resources/testdata/NEIS0140-subset.fasta')))
        path_query = Path(str(files('mist').joinpath('resources/testdata/query-perfect_hits.fasta')))
        data_out = minimap2utils.align(path_query, path_db)
        self.assertGreater(len(data_out), 0)

    def test_minimap2_query_with_cigar(self) -> None:
        """
        Tests minimap2 query function with Cigar string.
        :return: None
        """
        path_db = Path(str(files('mist').joinpath('resources/testdata/NEIS0140-subset.fasta')))
        path_query = Path(str(files('mist').joinpath('resources/testdata/query-perfect_hits.fasta')))
        data_out = minimap2utils.align(path_query, path_db, include_cigar=True)
        self.assertGreater(len(data_out), 0)

if __name__ == '__main__':
    initialize_logging()
    unittest.main()



import concurrent.futures
import json
import logging
import re
import shutil
from pathlib import Path
from typing import Any

from Bio import SeqIO

from mist.app import NAME_DB_INFO
from mist.app.loggers.logger import logger
from mist.app.utils import (
    clusterutils,
    dbutils,
    minimap2utils,
    sequenceutils,
)
from mist.app.utils.clustersplit import ClusterSplit
from mist.app.utils.dependencies import check_dependencies


class MistIndex:
    """
    Main class to index schemes using MiST.
    """

    def __init__(self, paths_fasta: list[Path], path_profiles: Path | None, cutoff: int = 80, debug: bool = False) -> None:
        """
        Initializes the index script.
        :param paths_fasta: List of input FASTA files
        :param path_profiles: Path to profiles file (optional)
        :param cutoff: Clustering % identity cutoff
        :param debug: Enable debug mode
        :return: None
        """
        check_dependencies(['CD-HIT', 'minimap2', 'nucmer'])
        self._paths_fasta = paths_fasta
        self._path_profiles = path_profiles
        self._loci = [p.name.replace('.fasta', '') for p in paths_fasta]
        self._cutoff = cutoff
        self._debug = debug

    @staticmethod
    def process_locus(path_fasta: Path, dir_out: Path, cutoff: int, debug: bool) -> dict[str, Any]:
        """
        Processes a single locus
        :param path_fasta: Input FASTA path
        :param dir_out: Output directory
        :param cutoff: Clustering cutoff
        :param debug: Enable debug mode
        :return: Database information
        """
        dir_out.mkdir(parents=True, exist_ok=True)

        # Convert the input FASTA file to standard format
        path_fasta_full = dir_out / path_fasta.name
        dbutils.reformat_fasta(path_fasta, path_fasta_full)

        # Parse the FASTA file
        seq_by_id = {}
        with open(path_fasta_full) as handle:
            for i, seq in enumerate(SeqIO.parse(handle, 'fasta')):
                seq_by_id[seq.id] = {'seq': seq, 'idx': i}
        logger.info(f'Parsed: {len(seq_by_id):,} sequences from {path_fasta.name}')

        # Create clustered FASTA
        path_fasta_clustered = clusterutils.cluster_fasta(path_fasta_full, dir_out, cutoff, threads=1, debug=debug)
        clusters = clusterutils.parse_cluster_from_file(path_fasta_clustered)

        # Calculate hashes
        orientation_by_seq_id = {
            member['seq_id']: member['ori'] if member['ori'] is not None else '+'
            for cluster in clusters for member in cluster['members']
        }
        data_hashes = sequenceutils.extract_hashes(path_fasta_full, orientation_by_seq_id)
        logging.info(f'Parsed {len(data_hashes)} hashes from {path_fasta.name}')

        # Output data
        data_out = {
            'allele_id_regex': r'_(\d+)$',
            'alleles': {
                re.search(sequenceutils.REGEX_ALLELE, id_).group(1): {
                    'idx': seq_info['idx'],
                    'length': len(seq_info['seq']),
                }
                for id_, seq_info in seq_by_id.items()
            },
            'clusters': {c['members'][0]['seq_id']: c for c in clusters},
            'cutoff': cutoff,
            'fasta_clustered': path_fasta_clustered.name,
            'fasta_full': path_fasta_full.name,
            'name': path_fasta.name.replace('.fasta', ''),
            'hashes': {
                h: allele_id
                for allele_id, h in data_hashes[['allele_id', 'hash']].itertuples(
                    index=False, name=None
                )
            },
            'nb_seqs': len(data_hashes)
        }

        with open(dir_out / 'mist_db.json', 'w') as handle:
            json.dump(data_out, handle, indent=2)

        # Split up misaligned clusters
        splitter = ClusterSplit(dir_out, debug=debug)
        splitter.run()

        # Return the DB information
        return data_out

    def _extract_repr_alleles(self, dir_out: Path, threads: int) -> None:
        """
        Extracts the representative alleles from all FASTA files.
        :param dir_out: Output directory
        :param threads: Nb. of threads
        :return: None
        """
        res_by_locus = {}
        try:
            with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
                future_to_locus = {
                    executor.submit(
                        MistIndex.process_locus, **{
                            'path_fasta': path_fasta,
                            'dir_out': dir_out / locus,
                            'cutoff': self._cutoff,
                            'debug': self._debug
                        },
                    ): locus for locus, path_fasta in zip(self._loci, self._paths_fasta)
                }

                for future in concurrent.futures.as_completed(future_to_locus):
                    locus = future_to_locus[future]
                    try:
                        res_by_locus[locus] = future.result()
                    except BaseException as err:
                        logging.error(f"Locus '{locus}' generated exception: {err}")
                        raise RuntimeError(err)
        except KeyboardInterrupt as err:
            logger.warning("Keyboard interrupt received. Cancelling all pending tasks.")
            for future in future_to_locus:
                future.cancel()
            raise err

    def create_index(self, dir_out: Path, threads: int = 4) -> None:
        """
        Creates the index in the target output directory.
        :param dir_out: Output directory
        :param threads: Number of threads
        :return: None
        """
        logger.info(f'Creating index for {len(self._paths_fasta):,} FASTA files')

        # Create the output directory
        dir_out = dir_out.expanduser().resolve()
        dir_out.mkdir(parents=True, exist_ok=True)

        # Extract representative alleles by clustering
        self._extract_repr_alleles(dir_out, threads)
        logger.info("Clustering completed successfully")

        # Create a FASTA file with all loci combined
        path_fasta_out = dir_out / 'loci_repr.fasta'
        nb_seqs = sequenceutils.merge_fasta_files(
            paths_fasta=[dir_out / locus / f'{locus}-clustered.fasta' for locus in self._loci],
            path_out=path_fasta_out,
        )
        minimap2utils.create_index(path_fasta_out)
        logger.info(f'Combined FASTA file created: {path_fasta_out} ({nb_seqs:,} sequences)')

        # Create a TXT file with all loci
        self._copy_db_info(dir_out)
        with open(dir_out / 'loci.txt', 'w') as handle:
            for locus in self._loci:
                handle.write(locus + '\n')

        # Copy profiles file
        if self._path_profiles:
            shutil.copyfile(self._path_profiles, dir_out / 'profiles.tsv')
        logger.info('Indexing finished successfully')

    def _copy_db_info(self, dir_out: Path) -> None:
        """
        Copies the database metadata information (if available).
        :param dir_out: Output directory
        :return: None
        """
        path_metadata = self._paths_fasta[0].parent / NAME_DB_INFO
        if not path_metadata.exists():
            logger.debug('No db_info.json file found, not copying DB metadata')
            return
        shutil.copyfile(path_metadata, dir_out / path_metadata.name)



from io import StringIO
from pathlib import Path

import pandas as pd

from mist.app.utils.command import Command


def nucmer(path_fasta_ref: Path, path_fasta_in: Path, dir_out: Path, threads: int = 1, debug: bool = False) -> Path:
    """
    Runs Minimap2 on the input sequence and DB.
    :param path_fasta_ref: Reference sequence FASTA file
    :param path_fasta_in: Input FASTA path
    :param dir_out: Output directory
    :param threads: Number of threads to use
    :param debug: If True, enable debug mode
    :return: None
    """
    basename = path_fasta_in.name.replace('.fasta', '')
    path_out = dir_out / basename
    command = Command(' '.join([
        'nucmer',
        '--maxmatch',
        f'-p {path_out}',
        str(path_fasta_ref),
        str(path_fasta_in),
        '--threads', str(threads) # nb. of threads
    ]))
    command.run(Path().cwd(), disable_logging=not debug)
    if not command.exit_code == 0:
        raise RuntimeError(f'Error running nucmer: {command.stderr}')
    return path_out.parent / f'{basename}.delta'


def show_coords(path_tsv: Path, debug: bool = False) -> pd.DataFrame:
    """
    Extracts the coordinates from the delta file.
    :param path_tsv: Path to delta file
    :param debug: If True, enable debug mode
    :return: Parsed coordinates
    """
    command = Command(' '.join([
        'show-coords',
        '-rcl',
        '-T',
        str(path_tsv)
    ]))
    command.run(Path().cwd(), disable_logging=not debug)
    if not command.exit_code == 0:
        raise RuntimeError(f'Error running show-coords: {command.stderr}')
    return pd.read_table(StringIO(command.stdout), skiprows=2)



import time
from pathlib import Path

import requests

from mist.app.loggers.logger import logger
from mist.app.utils import sequenceutils

REQUEST_SLEEP_SECONDS = 1


def retrieve_page_data(url: str, retries: int = 3, timeout: int = 20) -> requests.Response:
    """
    Retrieves data from the given URL.
    :param url: URL
    :param retries: Number of retries
    :param timeout: Timeout (in seconds)
    :return: URL data
    """
    error = None
    for retry in range(retries):
        try:
            response = requests.get(url, timeout=timeout)
            response.raise_for_status()
            return response
        except requests.exceptions.RequestException as err:
            logger.warning(f'Error retrieving {url} ({err!r}), retrying ({retry + 1}/{retries})')
            error = err
            time.sleep(min(10, 1 * (2**retry)))
    raise RuntimeError(f'Failed to retrieve {url} after {retries} attempts: {type(error).__name__}: {error}')


def download_fasta(locus_name: str, url: str, dir_out: Path) -> Path:
    """
    Downloads the FASTA for the given locus.
    :param locus_name: Locus name
    :param url: URL
    :param dir_out: Output directory
    :return: FASTA filename
    """
    response = retrieve_page_data(url)
    path_out = dir_out / f'{locus_name}.fasta'
    with path_out.open('w', encoding='utf-8') as handle:
        handle.write(response.text)
    logger.debug(f'FASTA file downloaded: {path_out.name} ({sequenceutils.count_sequences(path_out)})')
    time.sleep(REQUEST_SLEEP_SECONDS)
    return path_out



import logging
from pathlib import Path

logger = logging.getLogger('MiST')


def initialize_logging(log_path: Path | None = None, debug: bool = False) -> None:
    """
    Initializes the logging.
    :param log_path: Path to store the log
    :param debug: If true, enable debug mode
    :return: None
    """
    formatter = logging.Formatter('%(asctime)s - %(module)15s - %(levelname)7s - %(message)s')

    # Console handler
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    console_handler.setLevel(logging.INFO if not debug else logging.DEBUG)
    console_handler.name = 'console'
    logger.addHandler(console_handler)

    # File handler (pacu.log file)
    if log_path is not None:
        file_handler = logging.FileHandler(log_path)
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logging.DEBUG)
        logger.addHandler(file_handler)

    # General logging level
    logger.setLevel(logging.DEBUG)
    logger.propagate = False



import abc
import datetime
import json
from pathlib import Path
from typing import Any, ClassVar, Optional

from furl import furl

from mist.app import NAME_DB_INFO
from mist.app.loggers.logger import logger


class BaseDownloader(metaclass=abc.ABCMeta):
    """
    Baseclass for scheme downloaders.
    """
    DOWNLOADER_KEY: ClassVar[str]

    def __init__(self, **kwargs: Any) -> None:
        """
        Initializes the downloader.
        :param kwargs: Keyword arguments
        :return: None
        """
        self.dir_out: Optional[Path] = None

    @abc.abstractmethod
    def _download(self, url: str, dir_out: Path, include_profiles: bool = False) -> None:
        """
        Downloads the target scheme.
        :param url: Scheme URL
        :param dir_out: Output directory
        :param include_profiles: Include profiles
        :return: None
        """
        pass

    def download_scheme(self, url: str, dir_out: Path, include_profiles: bool = False) -> None:
        """
        Downloads the target scheme.
        :param url: Scheme URL
        :param dir_out: Output directory
        :param include_profiles: Include profiles
        :return: None
        """
        self.dir_out = dir_out
        self._download(url, dir_out, include_profiles)
        self.export_metadata_file(url)
        logger.info(
            f'You can create the index using:\n'
            f"mist index --fasta-list {dir_out / 'fasta_list.txt'} --output DB_NAME --threads 4")

    def create_fasta_list(self, paths_fasta: list[Path]) -> None:
        """
        Creates a TXT file with the FASTA file paths.
        :param paths_fasta: List of FASTA paths
        :return: None
        """
        path_fasta_list = self.dir_out / 'fasta_list.txt'
        with open(path_fasta_list, 'w') as handle:
            for path_fasta in paths_fasta:
                handle.write(str(path_fasta.absolute()))
                handle.write('\n')
        logger.info(f'FASTA list created: {path_fasta_list}')

    def export_metadata_file(self, url: str) -> None:
        """
        Exports the database metadata file.
        :param url: Scheme url
        :return: None
        """
        path_out = self.dir_out / NAME_DB_INFO
        with path_out.open('w') as handle:
            json.dump({
                'url': url,
                'downloader': self.DOWNLOADER_KEY,
                'download_date': datetime.datetime.now().isoformat(),
            }, handle, indent=2)
        logger.debug(f'DB info exported to: {path_out}')

    def get_available_schemes(self, base_url: furl, **kwargs: Any) -> list[tuple[str, str]]:
        """
        Retrieves the available schemes that can be downloaded.
        :param base_url: Base URL
        :param kwargs: Keyword arguments
        :return: List of available schemes
        """
        return []



import shutil
from importlib.resources import files
from pathlib import Path
from typing import Any

from furl import furl

from mist.app.dbs.basedownloader import BaseDownloader
from mist.app.loggers.logger import logger
from mist.app.utils import restutils, sequenceutils
from mist.app.utils.command import Command


class BIGSDbAuthDownloader(BaseDownloader):
    """
    Downloader for BIGSdb databases with authentication.
    """

    DOWNLOADER_KEY = 'bigsdb_auth'

    def __init__(self, **kwargs: Any) -> None:
        """
        Initializes the downloader.
        :param kwargs: Keyword arguments
        :return: None
        """
        super().__init__()
        try:
            self._dir_tokens: Path = Path(kwargs['dir_tokens'])
            self._key_name: str = kwargs['key_name']
            self._site: str = kwargs['site']
        except KeyError as err:
            raise RuntimeError(f"Required argument {err} missing")
        if self._dir_tokens is None:
            raise ValueError("Token directory should be set")

    @property
    def path_script(self) -> str:
        """
        Returns the path to the downloader script
        :return: Path to the downloader script
        """
        return str(files('mist').joinpath('resources/pubmlst/download_pubmlst.py'))

    def _check_credentials(self, dir_keys: Path) -> None:
        """
        Checks if the required credential are present.
        :param dir_keys: Path to the directory containing the credentials.
        :return: None
        """
        if not dir_keys.exists():
            raise FileNotFoundError(f'Directory does not exist: {dir_keys}')

    def _retrieve_page(self, url: str, path_out: Path) -> Path:
        """
        Downloads a FASTA file using OAuth.
        :param url: URL to the FASTA file
        :param path_out: Output path
        :return: None
        """
        command = Command(' '.join([
            'python', self.path_script,
            '--key_name', self._key_name,
            '--site', self._site,
            '--url', url,
            '--token_dir', str(self._dir_tokens.absolute()),
            '--cron',
            f'> {path_out.absolute()}'
        ]))
        command.run(self.dir_out, timeout=60)
        if not command.exit_code == 0:
            raise RuntimeError(f'Error retrieving: {url}\n{command.stderr}')
        return path_out

    def _download_fasta_files(self, metadata: dict[str, Any]) -> list[Path]:
        """
        Downloads the scheme FASTA files.
        :param metadata: Scheme metadata
        :return: List of FASTA paths
        """
        paths_fasta = []
        for href_locus in metadata['loci']:
            locus = furl(href_locus).path.segments[-1]
            url_fasta = str(furl(href_locus).add(path='alleles_fasta'))
            path_fasta = self._retrieve_page(url_fasta, self.dir_out / f'{locus}.fasta')
            paths_fasta.append(path_fasta)
            logger.info(
                f'Downloaded {path_fasta.name} ({sequenceutils.count_sequences(path_fasta):,} sequences)'
            )
        return paths_fasta

    def _download(self, url: str, dir_out: Path, include_profiles: bool = False) -> None:
        """
        Downloads the target scheme.
        :param url: Scheme URL
        :param dir_out: Output directory
        :param include_profiles: Include profiles
        :return: None
        """
        # Download the scheme metadata
        metadata = restutils.retrieve_page_data(url).json()
        logger.info(f"Metadata received: {metadata['locus_count']:,} loci")

        # Create a temporary directory to store the tokens (to avoid overwriting the original files)
        dir_tokens_temp = dir_out / 'tokens'
        dir_tokens_temp.mkdir(exist_ok=True)
        for p_file in self._dir_tokens.iterdir():
            if p_file.is_file():
                shutil.copyfile(p_file, dir_tokens_temp / p_file.name)
        logger.info(f"Tokens copied to: {dir_tokens_temp}")
        self._dir_tokens = dir_tokens_temp

        # Download FASTA files
        paths_fasta = self._download_fasta_files(metadata)

        # Download profiles
        if include_profiles:
            href_profiles = str(furl(url).add(path='profiles_csv'))
            self._retrieve_page(href_profiles, self.dir_out / 'profiles.tsv')
            logger.info('Profiles downloaded')

        # Create a TXT file with all FASTA files
        self.create_fasta_list(paths_fasta)
        logger.info('Scheme downloaded successfully')



import unittest

from click.testing import CliRunner

from mist.app.loggers.logger import logger
from mist.scripts.cli import cli
from mist.version import __version__


class TestVersion(unittest.TestCase):
    """
    Tests the version command.
    """

    def test_version(self) -> None:
        """
        Tests the version command.
        :return: None
        """
        runner = CliRunner()
        # noinspection PyTypeChecker
        result = runner.invoke(cli,['--version'])
        logger.info(result.output)
        self.assertIn(__version__, result.output)
        self.assertTrue(result.exit_code == 0)



from pathlib import Path

import pandas as pd
from Bio import SeqIO

from mist.app.loggers.logger import logger


class InvalidLengthException(Exception):
    """
    Error that is raised when the length of the input sequence does not match any sequences in the database.
    """

    def __init__(self, len_seq: int, allowed: list[int]) -> None:
        """
        Initializes the exception.
        :param len_seq: Sequence length
        :param allowed: Allowed sequence lengths
        """
        self.len_seq = len_seq
        self.allowed = allowed


class ImperfectMatchDetector:
    """
    Identifies the best matching imperfect hit.
    """

    def __init__(self, dir_in: Path) -> None:
        """
        Initializes the detector.
        :param dir_in: Input directory
        """
        self._dir_in = dir_in
        # Parse the database sequences
        records_in = []
        with (self._dir_in / f'{self._dir_in.name}.fasta').open() as handle:
            for seq in SeqIO.parse(handle, 'fasta'):
                records_in.append({
                    'id': seq.id,
                    'seq': str(seq.seq),
                    'length': len(seq)
                })
        self._data_seqs_db = pd.DataFrame(records_in)
        logger.debug(f'Parsed: {len(self._data_seqs_db):,} sequences ({dir_in.name})')

    def retrieve_best_matching(self, seq: str, min_id: int) -> list[str]:
        """
        Retrieves the best matching sequence(s) for the target sequence.
        :param seq: Target sequence
        :param min_id: Min. % sequence identity
        :return: Seq ids for the best matching sequences
        """
        data_subset = self._data_seqs_db[self._data_seqs_db['length'] == len(seq)].copy()
        logger.debug(f'Found {len(data_subset):,} allele(s) matching the length of the detected sequence ({len(seq)})')
        if len(data_subset) == 0:
            viable_lengths = list(self._data_seqs_db['length'].unique())
            logger.debug(
                f"Length of detected sequence ({len(seq):,}) does not match any alleles in the "
                f"database ({', '.join(str(l) for l in sorted(viable_lengths))})")
            raise InvalidLengthException(len(seq), viable_lengths)

        data_subset['nb_matches'] = data_subset['seq'].apply(lambda x: sum(c1 == c2 for c1, c2 in zip(x, seq)))
        max_matches = data_subset['nb_matches'].max()

        # Check if the identity matches
        identity = 100 * max_matches / len(seq)
        if identity <= min_id:
            logger.debug(f'Identity ({identity:.2f}%) to best matching sequence is below threshold ({min_id}%).')
            return []
        return list(data_subset[data_subset['nb_matches'] == max_matches]['id'])



import dataclasses
import json
from datetime import datetime
from pathlib import Path

import pandas as pd
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from mist.app import model, NAME_DB_INFO
from mist.app.loggers.logger import logger
from mist.app.model import CustomEncoder
from mist.app.query.allelequeryminimap import AlleleQueryMinimap2, MultiStrategy
from mist.app.query.profilequery import ProfileQuery
from mist.app.utils.dependencies import check_dependencies
from mist.version import __version__


class MistCaller:
    """
    Main class to call alleles.
    """

    def __init__(self, dir_db: Path, multi: str, loci: list[str] | None = None, keep_minimap2: bool = False,
                 export_novel: bool = False, min_id_novel: int = 99) -> None:
        """
        Initializes a query.
        :param dir_db: Database directory
        :param multi: Multi-hit strategy
        :param loci: Restrict typing to these loci
        :param keep_minimap2: Keep Minimap2 output
        :param export_novel: If True, novel alleles are exported in FASTA format
        :param min_id_novel: Minimum sequence identity for novel alleles
        :return: None
        """
        check_dependencies(['minimap2'])
        self._dir_db = dir_db
        self._multi = multi
        self._loci = loci
        self._keep_minimap2 = keep_minimap2
        self._export_novel = export_novel
        self._min_id_novel = min_id_novel

    def _results_to_df(self, result_by_locus: dict[str, model.QueryResult]) -> pd.DataFrame:
        """
         Converts the typing results into a dataframe.
        :param result_by_locus: Results by locus.
        :return: DataFrame
        """
        data_out = pd.DataFrame([{
            # Main result
            'locus': locus,
            'allele': res.allele_str,
            'length': ';'.join(str(x.length) for x in res.allele_results) if len(res.allele_results) > 0 else '-',

            # Location in genome
            'contig': ';'.join(x.alignment.seq_id for x in res.allele_results) if len(res.allele_results) > 0 else '-',
            'start': ';'.join(str(x.alignment.start) for x in res.allele_results) if len(res.allele_results) > 0 else '-',
            'end': ';'.join(str(x.alignment.end) for x in res.allele_results) if len(res.allele_results) > 0 else '-',
            'strand': ';'.join(str(x.alignment.strand) for x in res.allele_results) if len(res.allele_results) > 0 else '-',

            # Novel alleles
            'is_novel': model.Tag.NOVEL in res.tags,
            'closest_alleles': ';'.join(res.allele_results[0].closest_alleles) if model.Tag.NOVEL in res.tags else '-',

            # Additional tags
            'tags': ';'.join(t.value for t in res.tags) if len(res.tags) > 0 else '-'}
            for locus, res in result_by_locus.items()
        ])

        # noinspection PyTypeChecker,PyUnresolvedReferences
        nb_detected = (data_out['allele'] != '-').sum()
        nb_novel = data_out['is_novel'].sum()
        logger.info(
            f"Detected {nb_detected}/{len(data_out)} loci ({100 * nb_detected / len(data_out):.2f}%), "
            f"including {nb_novel:,} (potential) novel alleles"
        )
        return data_out

    def _export_novel_allele_seqs(
            self, data_results: pd.DataFrame, res_by_locus: dict[str, model.QueryResult], dir_out: Path) -> None:
        """
        Exports the novel allele sequences.
        :param data_results: DataFrame with typing results
        :param res_by_locus: Results by locus
        :param dir_out: Output directory
        :return: None
        """
        for locus in data_results[data_results['is_novel']]['locus']:
            dir_novel_alleles = dir_out / 'novel_alleles'
            dir_novel_alleles.mkdir(parents=True, exist_ok=True)
            allele_id = res_by_locus[locus].allele_str.replace('*', '')
            with open(dir_novel_alleles / f'{locus}-{allele_id}.fasta', 'w') as handle:
                SeqIO.write(SeqRecord(
                    id=f'{locus}_{allele_id}',
                    description=f"closest_match={','.join(res_by_locus[locus].allele_results[0].closest_alleles)}",
                    seq=Seq(res_by_locus[locus].allele_results[0].sequence),
                ), handle, 'fasta')

    def call_alleles(self, path_fasta: Path, out_json: Path, out_dir: Path, out_tsv: Path, threads: int) -> None:
        """
        Calls the alleles from the input FASTA file.
        :param path_fasta: Input FASTA path
        :param out_json: JSON output path
        :param out_dir: Output directory
        :param out_tsv: Output TSV path
        :param threads: Number of threads
        :return: None
        """
        allele_query = AlleleQueryMinimap2(
            dir_db=self._dir_db,
            dir_out=out_dir,
            multi_strategy=MultiStrategy(self._multi),
            min_id_novel=self._min_id_novel,
            save_minimap2=self._keep_minimap2,
        )
        result_by_locus = allele_query.query(path_fasta, loci=self._loci, threads=threads)

        # Create DataFrame to calculate statistics
        data_results = self._results_to_df(result_by_locus)

        # Query the profiles
        if (self._dir_db / 'profiles.tsv').exists():
            profile_query = ProfileQuery(self._dir_db / 'profiles.tsv')
            profile, nb_matches = profile_query.query(result_by_locus)
            pct_match = 100 * nb_matches / len(result_by_locus)
            logger.info(f'Matching ST: {profile.name} ({pct_match:.2f}% match)')
        else:
            profile = None
            pct_match = None
            nb_matches = None

        # Create output files
        self._export_json(result_by_locus, path_out=out_json, profile=profile, nb_matches=nb_matches, pct_match=pct_match)
        if self._export_novel and data_results['is_novel'].sum() > 0:
            logger.info("Exporting novel allele sequences")
            self._export_novel_allele_seqs(data_results, result_by_locus, out_dir)

        # Export TSV output
        if out_tsv is not None:
            data_results.to_csv(out_tsv, sep='\t', index=False)

    def _export_json(
            self, results_by_locus: dict[str, model.QueryResult], path_out: Path, profile: model.Profile,
            nb_matches: int | None, pct_match: float | None) -> None:
        """
        Exports the results in JSON format.
        :param results_by_locus: Result(s) by locus
        :param path_out: Output path
        :param profile: Detected profile
        :param nb_matches: Number of matching loci
        :param pct_match: Percent match for the profile
        :return: None
        """
        # Parse the database information (if available)
        path_db_info = self._dir_db / NAME_DB_INFO
        if path_db_info.exists():
            with path_db_info.open() as handle:
                data_db = json.load(handle)
        else:
            data_db = None

        with open(path_out, 'w') as handle:
            json.dump({
                'alleles': {locus: dataclasses.asdict(res) for locus, res in results_by_locus.items()},
                'profile': {
                    **dataclasses.asdict(profile),
                    'pct_match': pct_match,
                    'nb_matches': nb_matches
                } if profile is not None else None,
                'metadata': {
                    'timestamp': datetime.now().isoformat(),
                    'tool_version': __version__,
                    'db': data_db
                }
            }, handle, indent=2, cls=CustomEncoder)
        logger.info(f'Output stored in: {path_out}')



from pathlib import Path

from mist.app.errors import DependencyError
from mist.app.loggers.logger import logger
from mist.app.utils.command import Command

VERSION_COMMANDS = {
    'CD-HIT': {'command': 'cd-hit -h', 'exit_code': 1},
    'minimap2': {'command': 'minimap2 --version', 'exit_code': 0},
    'nucmer': {'command': 'nucmer --version', 'exit_code': 0},
}

def _check_dependency(key: str) -> None:
    """
    Checks if the given dependency is available.
    :param key: Dependency key
    :return: None
    """
    if key not in VERSION_COMMANDS:
        raise ValueError(f'Unknown dependency key: {key}')
    command = Command(VERSION_COMMANDS[key]['command'])
    command.run(Path.cwd(), disable_logging=True)
    if command.exit_code != VERSION_COMMANDS[key]['exit_code']:
        raise DependencyError(f"Dependency '{key}' is not available")

def check_dependencies(keys: list[str]) -> None:
    """
    Checks if the given dependencies are available.
    :param keys: Keys to check
    :return: None
    """
    logger.info('Checking if dependencies are available')

    for key in keys:
        _check_dependency(key)
        logger.info(f'  {key}: OK')



import unittest
from importlib.resources import files
from pathlib import Path

import pandas as pd
from click.testing import CliRunner

from mist.app.utils import testingutils
from mist.scripts.cli import cli


class TestDists(unittest.TestCase):
    """
    Tests for the mist dists script.
    """

    @staticmethod
    def get_output_files(ext: str) -> list[Path]:
        """
        Returns the output files for testing with the given extension.
        :param ext: Extension of the output files.
        :return: List of files
        """
        dir_test = Path(str(files('mist').joinpath('resources/testdata/output')))
        return sorted(list(dir_test.glob(f'*{ext}')))

    def _is_valid_output_file(self, path: Path) -> bool:
        """
        Checks if the (tabular) output file is valid.
        :param path: Output file
        :return: True if valid
        """
        df_matrix = pd.read_table(path)
        self.assertFalse(df_matrix.empty)
        return True

    def test_tsv_input(self) -> None:
        """
        Tests the mist dists script with TSV input.
        :return: None
        """
        runner = CliRunner()
        with testingutils.get_temp_dir() as dir_temp:
            path_out_dists = Path(dir_temp, 'output_dists.tsv')
            path_out_matrix = Path(dir_temp, 'output_matrix.tsv')
            # noinspection PyTypeChecker
            result = runner.invoke(cli, [
                'dists',
                *(str(x) for x in TestDists.get_output_files('.tsv')),
                '--out-dists', str(path_out_dists),
                '--out-matrix', str(path_out_matrix)
            ], catch_exceptions=False)

            # Verify output
            self.assertTrue(self._is_valid_output_file(path_out_dists))
            self.assertTrue(self._is_valid_output_file(path_out_matrix))
            self.assertEqual(result.exit_code, 0)

    def test_json_input(self) -> None:
        """
        Tests the mist dists script with JSON input.
        :return: None
        """
        runner = CliRunner()
        with testingutils.get_temp_dir() as dir_temp:
            path_out_dists = Path(dir_temp, 'output_dists.tsv')
            path_out_matrix = Path(dir_temp, 'output_matrix.tsv')
            # noinspection PyTypeChecker
            result = runner.invoke(cli, [
                'dists',
                *(str(x) for x in TestDists.get_output_files('.json')),
                '--out-dists', str(path_out_dists),
                '--out-matrix', str(path_out_matrix)
            ])

            # Verify output
            self.assertTrue(self._is_valid_output_file(path_out_dists))
            self.assertTrue(self._is_valid_output_file(path_out_matrix))
            self.assertEqual(result.exit_code, 0)

    def test_mixed_input(self) -> None:
        """
        Tests the mist dists script with JSON input.
        :return: None
        """
        runner = CliRunner()
        with testingutils.get_temp_dir() as dir_temp:
            path_out_dists = Path(dir_temp, 'output_dists.tsv')
            path_out_matrix = Path(dir_temp, 'output_matrix.tsv')
            # noinspection PyTypeChecker
            result = runner.invoke(cli, [
                'dists',
                *(str(x) for x in TestDists.get_output_files('.json')[:2]),
                *(str(x) for x in TestDists.get_output_files('.tsv')[2:]),
                '--out-dists', str(path_out_dists),
                '--out-matrix', str(path_out_matrix)
            ])

            # Verify output
            self.assertTrue(self._is_valid_output_file(path_out_dists))
            self.assertTrue(self._is_valid_output_file(path_out_matrix))
            self.assertEqual(result.exit_code, 0)

    def test_invalid_input(self) -> None:
        """
        Tests the mist dists script with invalid input (too few input files).
        :return: None
        """
        runner = CliRunner()
        with testingutils.get_temp_dir() as dir_temp:
            path_out_dists = Path(dir_temp, 'output_dists.tsv')
            path_out_matrix = Path(dir_temp, 'output_matrix.tsv')
            # noinspection PyTypeChecker
            result = runner.invoke(cli, [
                'dists',
                next(str(x) for x in TestDists.get_output_files('.json')),
                '--out-dists', str(path_out_dists),
                '--out-matrix', str(path_out_matrix)
            ])
            self.assertNotEqual(result.exit_code, 0)


if __name__ == '__main__':
    unittest.main()



from pathlib import Path

import pandas as pd

from mist.app import model
from mist.app.loggers.logger import logger


class ProfileQuery:
    """
    Class to query profiles files with a given allele combination.
    """

    ALLELE_ABSENT = '0'
    ALLELE_WILDCARD = 'N'

    def _parse_profiles(self, path: Path, locus_names: set[str]) -> list[model.Profile]:
        """
        Parses the profile file.
        :param path: Path to the TSV file
        :param locus_names: Locus names
        :return: List of profiles
        """
        # Parse input data
        data_in = pd.read_table(path, dtype=str)
        cols_metadata = [c for c in data_in.columns if c not in locus_names]
        logger.debug(f'Metadata columns: {cols_metadata}')
        cols_alleles = [c for c in data_in.columns if c in locus_names]
        logger.debug(f'Gene columns: {cols_alleles}')

        # Construct the profiles
        profiles = []
        for row in data_in.fillna('n/a').to_dict('records'):
            profiles.append(model.Profile(
                name=row[data_in.columns[0]],
                alleles={c: row[c] for c in cols_alleles},
                metadata=[(c, row[c] if not pd.isna(row[c]) else '-') for c in cols_metadata],
            ))
        logger.debug(f'Parsed {len(profiles):,} profiles')
        return profiles

    @staticmethod
    def _alleles_match(res: model.QueryResult, profile_allele: str) -> bool:
        """
        Checks whether two alleles match.
        :param res: Result
        :param profile_allele: Profile allele
        :return: True if the alleles match, False otherwise
        """
        if profile_allele == ProfileQuery.ALLELE_WILDCARD:
            return True
        if profile_allele == ProfileQuery.ALLELE_ABSENT:
            return res is None

        # No allele detected
        if res is None:
            return False

        # Single allele detected
        if len(res.allele_results) == 1:
            return res.allele_str == profile_allele

        # Multiple alleles detected (-> match if one of them matches)
        return any(allele == profile_allele for allele in res.allele_str.split('__'))

    def __init__(self, path_profiles: Path) -> None:
        """
        Initializes the profiles query class.
        :param path_profiles: Path to profiles file
        :return: None
        """
        locus_names = set(d.name for d in path_profiles.parent.iterdir() if not d.name.startswith('.'))
        self._profiles_by_name = {p.name: p for p in self._parse_profiles(path_profiles, locus_names)}

    def query(self, result_by_locus: dict[str, model.QueryResult | None]) -> tuple[model.Profile, int]:
        """
        Queries the profiles using the detected alleles.
        :param result_by_locus: Detected allele(s) by locus
        :return: Best matching profile, nb. of matching loci
        """
        data_profiles = pd.DataFrame({'profile': [name for name in self._profiles_by_name.keys()]})
        data_profiles['nb_matches'] = data_profiles['profile'].apply(
            lambda p: sum(
                ProfileQuery._alleles_match(res, self._profiles_by_name[p].alleles[locus])
                for locus, res in result_by_locus.items()))
        best_match = data_profiles.sort_values(by='nb_matches', ascending=False).iloc[0]
        return self._profiles_by_name[best_match['profile']], int(best_match['nb_matches'])



import os
import unittest
from pathlib import Path

from mist.app.dbs.bigsdbauthdownloader import BIGSDbAuthDownloader
from mist.app.dbs.bigsdbdownloader import BIGSDbDownloader
from mist.app.dbs.cgmlstorgdownloader import CgMLSTOrgDownloader
from mist.app.dbs.enterobasedownloader import EnteroBaseDownloader
from mist.app.loggers.logger import initialize_logging
from mist.app.utils import testingutils


class TestDownloaders(unittest.TestCase):
    """
    Tests for the downloading MLST schemes.
    """

    @staticmethod
    def download_succeeded(dir_scheme: Path, with_profiles: bool) -> bool:
        """
        Checks if the scheme in the target directory is valid.
        :param dir_scheme: Directory to check
        :param with_profiles: If true, profiles are checked
        :return: True if valid
        """
        path_fasta_list = dir_scheme / 'fasta_list.txt'
        if not path_fasta_list.exists():
            raise FileNotFoundError(f'{path_fasta_list.name} is missing')
        with path_fasta_list.open() as handle:
            for p in [Path(x.strip()) for x in handle.readlines()]:
                if not p.exists():
                    raise FileNotFoundError(f'{p.name} is missing')
        if with_profiles:
            if not (dir_scheme / 'profiles.tsv').exists():
                raise FileNotFoundError('profiles.tsv file is missing')
        return True

    def setUp(self) -> None:
        """
        Sets up a temporary directory and builds a database there before each test.
        :return: None
        """
        self.dir_temp = testingutils.get_temp_dir()

    def tearDown(self) -> None:
        """
        Clean up the temporary directory after the test.
        :return: None
        """
        self.dir_temp.cleanup()

    def test_download_cgmlst_org(self) -> None:
        """
        Tests downloading a scheme from cgMLST.org.
        :return: None
        """
        with testingutils.get_temp_dir() as dir_temp:
            downloader = CgMLSTOrgDownloader()
            downloader.download_scheme(
                url='https://www.cgmlst.org/ncs/schema/Cjejuni22/',
                dir_out=Path(dir_temp),
                include_profiles=True)
            self.assertTrue(TestDownloaders.download_succeeded(Path(dir_temp), with_profiles=True))

    def test_download_enterobase(self) -> None:
        """
        Tests downloading a scheme from EnteroBase.
        :return: None
        """
        with testingutils.get_temp_dir() as dir_temp:
            downloader = EnteroBaseDownloader()
            downloader.download_scheme(
                url='https://enterobase.warwick.ac.uk/schemes/Yersinia.Achtman7GeneMLST/',
                dir_out=Path(dir_temp),
                include_profiles=True)
            self.assertTrue(TestDownloaders.download_succeeded(Path(dir_temp), with_profiles=True))

    def test_download_pubmlst(self) -> None:
        """
        Tests downloading a scheme from PubMLST (without authentication).
        :return: None
        """
        with testingutils.get_temp_dir() as dir_temp:
            downloader = BIGSDbDownloader()
            downloader.download_scheme(
                url='https://rest.pubmlst.org/db/pubmlst_saureus_seqdef/schemes/1',
                dir_out=Path(dir_temp),
                include_profiles=True)
            self.assertTrue(TestDownloaders.download_succeeded(Path(dir_temp), with_profiles=True))

    @unittest.skipUnless(os.environ.get('MIST_AUTH'), "Skipping test (MIST_AUTH not set).")
    def test_download_pubmlst_with_auth(self) -> None:
        """
        Tests downloading a scheme from PubMLST (with authentication).
        :return: None
        """
        with testingutils.get_temp_dir() as dir_temp:
            downloader = BIGSDbAuthDownloader(
                site='PubMLST',
                key_name='PubMLST',
                dir_tokens=Path(os.environ.get('MIST_AUTH'))
            )
            downloader.download_scheme(
                url='https://rest.pubmlst.org/db/pubmlst_saureus_seqdef/schemes/1',
                dir_out=Path(dir_temp),
                include_profiles=True
            )
            self.assertTrue(TestDownloaders.download_succeeded(Path(dir_temp), with_profiles=True))


if __name__ == '__main__':
    initialize_logging()
    unittest.main()



#!/usr/bin/env python3
# Script to download authenticated resources from PubMLST and BIGSdb Pasteur
# via their REST interfaces.
# Written by Keith Jolley
# Copyright (c) 2024, University of Oxford
# E-mail: keith.jolley@biology.ox.ac.uk
#
# BIGSdb_downloader is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BIGSdb_downloader is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# Version 20241030
import argparse
import os
import stat
import re
import configparser
import sys
import json
from urllib.parse import parse_qs
from pathlib import Path
from rauth import OAuth1Service, OAuth1Session

BASE_WEB = {
    "PubMLST": "https://pubmlst.org/bigsdb",
    "Pasteur": "https://bigsdb.pasteur.fr/cgi-bin/bigsdb/bigsdb.pl",
}
BASE_API = {
    "PubMLST": "https://rest.pubmlst.org",
    "Pasteur": "https://bigsdb.pasteur.fr/api",
}


parser = argparse.ArgumentParser()

parser.add_argument(
    "--cron",
    action="store_true",
    help="Script is being run as a CRON job or non-interactively.",
)
parser.add_argument(
    "--db", required=False, help="Database config - only needed for setup."
)
parser.add_argument(
    "--json_body",
    required=False,
    help="JSON body to be included in a POST call. If this is longer than "
    "the command line limit (probably about 128kb) then you will need to "
    "save the JSON payload to a file and use --json_body_file",
)
parser.add_argument(
    "--json_body_file",
    required=False,
    help="File containing JSON to use in the body of a POST call.",
)
parser.add_argument(
    "--key_name",
    required=True,
    help="Name of API key - use a different name for each site.",
)
parser.add_argument(
    "--method",
    required=False,
    choices=["GET", "POST"],
    default="GET",
    help="HTTP method",
)
parser.add_argument(
    "--output_file",
    required=False,
    help="Path and filename of saved file. Output sent to STDOUT if not specified.",
)
parser.add_argument(
    "--setup", action="store_true", help="Initial setup to obtain access token."
)
parser.add_argument("--site", required=False, choices=["PubMLST", "Pasteur"])
parser.add_argument(
    "--token_dir",
    required=False,
    default="./.bigsdb_tokens",
    help="Directory into which keys and tokens will be saved.",
)
parser.add_argument("--url", required=False, help="URL for API call.")
args = parser.parse_args()


def main():
    check_required_args(args)
    check_dir(args.token_dir)
    if args.setup:
        (access_token, access_secret) = get_new_access_token()
        if not access_token or not access_secret:
            raise PermissionError("Cannot get new access token.")
    (token, secret) = retrieve_token("session")
    if not token or not secret:
        (token, secret) = get_new_session_token()
    if not args.setup:
        get_route(args.url, token, secret)


def check_required_args(args):
    if args.setup:
        if not args.site:
            parser.error("--site is required for setup")
        if not args.db:
            parser.error("--db is required for setup")
    else:
        if not args.url:
            parser.error("--url is required")
    if args.json_body:
        if args.method != "POST":
            parser.error("You cannot use --json_body with --method=GET")
        if args.json_body_file:
            parser.error("You cannot use both --json_body and --json_body_file")

    if args.json_body_file:
        if args.method != "POST":
            parser.error("You cannot use --json_body_file with --method=GET")


def is_valid_json(json_string):
    try:
        json.loads(json_string)
        return True
    except ValueError:
        return False


def trim_url_args(url):
    if not "?" in url:
        return url, {}
    trimmed_url, param_string = url.split("?")
    params = parse_qs(param_string)
    params = {k: int(v[0]) for k, v in params.items()}

    return trimmed_url, params


def get_route(url, token, secret):
    (client_key, client_secret) = get_client_credentials()
    session = OAuth1Session(
        client_key, client_secret, access_token=token, access_token_secret=secret
    )
    trimmed_url, request_params = trim_url_args(url)
    if args.method == "GET":
        r = session.get(
            trimmed_url,
            params=request_params,
            headers={"User-Agent": "BIGSdb downloader"},
        )
    else:
        json_body = "{}"
        if args.json_body:
            json_body = args.json_body
        elif args.json_body_file:
            with open(args.json_body_file, "r") as file:
                json_body = file.read()
        if not is_valid_json(json_body):
            parser.error("Body does not contain valid JSON")
        r = session.post(
            trimmed_url,
            params=request_params,
            data=json_body,
            headers={
                "Content-Type": "application/json",
                "User-Agent": "BIGSdb downloader",
            },
            header_auth=True,
        )

    if r.status_code == 200 or r.status_code == 201:
        if args.output_file:
            try:
                with open(args.output_file, "w") as file:
                    if re.search("json", r.headers["content-type"], flags=0):
                        file.write(str(r.json()))
                    else:
                        file.write(r.text)
            except IOError as e:
                sys.stderr.write(f"An error occurred while writing to the file: {e}\n")
        else:
            if re.search("json", r.headers["content-type"], flags=0):
                print(r.json())
            else:
                print(r.text)
    elif r.status_code == 400:
        sys.stderr.write("Bad request")
        sys.stderr.write(r.json()["message"])
        sys.exit(1)
    elif r.status_code == 401:
        if re.search("unauthorized", r.json()["message"]):
            sys.stderr.write("Access denied - client is unauthorized\n")
            sys.exit(1)
        else:
            sys.stderr.write(r.json()["message"] + "\n")
            sys.stderr.write("Invalid session token, requesting new one...\n")
            (token, secret) = get_new_session_token()
            get_route(url, token, secret)
    else:
        sys.stderr.write(f"Error: {r.text}\n")
        sys.exit(1)


def check_dir(directory):
    if os.path.isdir(directory):
        if os.access(directory, os.W_OK):
            return
        else:
            raise PermissionError(
                f"The token directory '{directory}' exists but is not writable."
            )
    else:
        try:
            os.makedirs(directory)
            os.chmod(directory, stat.S_IRWXU)  # Set permissions to 0700
        except OSError as e:
            raise PermissionError(
                f"Failed to create token directory '{directory}': {e}"
            )


def retrieve_token(token_type):
    file_path = Path(f"{args.token_dir}/{token_type}_tokens")
    if file_path.is_file():
        config = configparser.ConfigParser(interpolation=None)
        config.read(file_path)
        if config.has_section(args.key_name):
            token = config[args.key_name]["token"]
            secret = config[args.key_name]["secret"]
            return (token, secret)
    return (None, None)


def get_new_session_token():
    file_path = Path(f"{args.token_dir}/session_tokens")
    (access_token, access_secret) = retrieve_token("access")
    if not access_token or not access_secret:
        (access_token, access_secret) = get_new_access_token()
    service = get_service()
    (client_key, client_secret) = get_client_credentials()
    db = get_db_value()
    url = f"{BASE_API[args.site]}/db/{db}/oauth/get_session_token"
    session_request = OAuth1Session(
        client_key,
        client_secret,
        access_token=access_token,
        access_token_secret=access_secret,
    )
    r = session_request.get(url, headers={"User-Agent": "BIGSdb downloader"})
    if r.status_code == 200:
        token = r.json()["oauth_token"]
        secret = r.json()["oauth_token_secret"]
        config = configparser.ConfigParser(interpolation=None)
        if file_path.is_file():
            config.read(file_path)
        config[args.key_name] = {"token": token, "secret": secret}
        with open(file_path, "w") as configfile:
            config.write(configfile)

        return (token, secret)
    else:
        sys.stderr.write(
            "Failed to get new session token. " + r.json()["message"] + "\n"
        )
        if args.cron:
            sys.stderr.write("Run interactively to fix.\n")
        if re.search("verification", r.json()["message"]) or re.search(
            "Invalid access token", r.json()["message"]
        ):
            sys.stderr.write("New access token required - removing old one.\n")
            config = configparser.ConfigParser(interpolation=None)
            file_path = Path(f"{args.token_dir}/access_tokens")
            if file_path.is_file():
                config.read(file_path)
                config.remove_section(args.key_name)
                with open(file_path, "w") as configfile:
                    config.write(configfile)
        sys.exit(1)


def get_service():
    db = get_db_value()
    (client_key, client_secret) = get_client_credentials()
    request_token_url = f"{BASE_API[args.site]}/db/{db}/oauth/get_request_token"
    access_token_url = f"{BASE_API[args.site]}/db/{db}/oauth/get_access_token"
    return OAuth1Service(
        name="BIGSdb_downloader",
        consumer_key=client_key,
        consumer_secret=client_secret,
        request_token_url=request_token_url,
        access_token_url=access_token_url,
        base_url=BASE_API[args.site],
    )


def get_new_access_token():
    if args.cron:
        sys.stderr.write(f"No access token saved for {args.key_name}.\n")
        sys.stderr.write("Run interactively to set.\n")
        sys.exit(1)
    file_path = Path(f"{args.token_dir}/access_tokens")
    (request_token, request_secret) = get_new_request_token()
    db = get_db_value()
    print(
        "Please log in using your user account at "
        f"{BASE_WEB[args.site]}?db={db}&page=authorizeClient&oauth_token={request_token} "
        "using a web browser to obtain a verification code."
    )
    verifier = input("Please enter verification code: ")
    service = get_service()
    r = service.get_raw_access_token(
        request_token,
        request_secret,
        params={"oauth_verifier": verifier},
        headers={"User-Agent": "BIGSdb downloader"},
    )
    if r.status_code == 200:
        token = r.json()["oauth_token"]
        secret = r.json()["oauth_token_secret"]
        file_path = Path(f"{args.token_dir}/access_tokens")
        print("Access Token:        " + token)
        print("Access Token Secret: " + secret + "\n")
        print(
            "This access token will not expire but may be revoked by the \n"
            f"user or the service provider. It will be saved to \n{file_path}."
        )
        config = configparser.ConfigParser(interpolation=None)
        if file_path.is_file():
            config.read(file_path)
        config[args.key_name] = {"token": token, "secret": secret}
        with open(file_path, "w") as configfile:
            config.write(configfile)
        return (token, secret)
    else:
        sys.stderr.write("Failed to get new access token." + r.json()["message"])
        sys.exit(1)


def get_db_value():
    db = None
    if args.db:
        db = args.db
    elif args.url:
        match = re.search(r"/db/([^/]+)", args.url)
        if match:
            db = match.group(1)
        else:
            raise ValueError("No db value found in the URL.")
    return db


def get_new_request_token():
    (client_key, client_secret) = get_client_credentials()
    db = get_db_value()
    service = get_service()

    r = service.get_raw_request_token(
        params={"oauth_callback": "oob"}, headers={"User-Agent": "BIGSdb downloader"}
    )
    if r.status_code == 200:
        token = r.json()["oauth_token"]
        secret = r.json()["oauth_token_secret"]
        return (token, secret)
    else:
        sys.stderr.write("Failed to get new request token." + r.json()["message"])
        sys.exit(1)


def get_client_credentials():
    config = configparser.ConfigParser(interpolation=None)
    file_path = Path(f"{args.token_dir}/client_credentials")
    client_id = None
    client_secret = None
    if file_path.is_file():
        config.read(file_path)
        if config.has_section(args.key_name):
            client_id = config[args.key_name]["client_id"]
            client_secret = config[args.key_name]["client_secret"]
    if not client_id:
        if args.cron:
            sys.stderr.write(f"No client credentials saved for {args.key_name}.\n")
            sys.stderr.write("Run interactively to set.\n")
            sys.exit(1)
        client_id = input("Enter client id: ").strip()
        while len(client_id) != 24:
            print("Client ids are exactly 24 characters long.")
            client_id = input("Enter client id: ").strip()
        client_secret = input("Enter client secret: ").strip()
        while len(client_secret) != 42:
            print("Client secrets are exactly 42 characters long.")
            client_secret = input("Enter client secret: ").strip()

        config[args.key_name] = {"client_id": client_id, "client_secret": client_secret}
        with open(file_path, "w") as configfile:
            config.write(configfile)
    return client_id, client_secret


if __name__ == "__main__":
    main()



import dataclasses
import json
from enum import Enum
from typing import Any

ALLELE_MISSING = '-'

class CustomEncoder(json.JSONEncoder):
    """
    Custom encoder that can handle dumping dataclass objects.
    """
    def default(self, obj: Any) -> Any:
        """
        Default encoding method.
        :param obj: Input object
        :return: Encoded object
        """
        if isinstance(obj, Enum):
            return obj.value
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
        return super().default(obj)

class Tag(Enum):
    """
    Tags to denote the different result types.
    """
    EXACT = 'EXACT' # Exact match
    NOVEL = 'NOVEL' # Potential novel allele
    MULTI = 'MULTI' # Multiple perfect hits
    ABSENT = 'ABSENT' # No seed alignment
    INDEL = 'INDEL' # Incorrect length
    EDGE = 'EDGE' # Contig edge


@dataclasses.dataclass(frozen=True, unsafe_hash=True)
class Profile:
    """
    Dataclass to hold ST profile data.
    """
    name: str
    metadata: list[tuple[str, str]] = dataclasses.field(hash=False)
    alleles: dict[str, str] | None = dataclasses.field(hash=False)


@dataclasses.dataclass(frozen=True)
class Alignment:
    """
    Represents an alignment in the input sequence.
    """
    seq_id: str
    start: int
    end: int
    strand: str

    @property
    def length(self) -> int:
        """
        Returns the length of the alignment.
        :return: Alignment length
        """
        return self.end - self.start


@dataclasses.dataclass(frozen=True)
class AlleleResult:
    """
    Match to an allele.
    """
    allele: str
    alignment: Alignment
    length: int
    sequence: str | None = None
    closest_alleles: list[str] | None = None


@dataclasses.dataclass(frozen=True)
class QueryResult:
    """
    Combined output of a locus query, can contain multiple allele matches
    """
    allele_str: str
    allele_results: list[AlleleResult]
    tags: list[Tag]



from mist.app.dbs.bigsdbauthdownloader import BIGSDbAuthDownloader
from mist.app.dbs.bigsdbdownloader import BIGSDbDownloader
from mist.app.dbs.cgmlstorgdownloader import CgMLSTOrgDownloader
from mist.app.dbs.enterobasedownloader import EnteroBaseDownloader

DOWNLOADERS = {
    'cgmlstorg': CgMLSTOrgDownloader,
    'enterobase': EnteroBaseDownloader,
    'bigsdb': BIGSDbDownloader,
    'bigsdb_auth': BIGSDbAuthDownloader,
}

BASE_URL_BY_DB = {
    'cgmlstorg': {'base': 'https://www.cgmlst.org/ncs'},
    'enterobase': {'base': 'https://enterobase.warwick.ac.uk/schemes'},
    'bigsdb': {'pubmlst': 'https://rest.pubmlst.org', 'pasteur': 'https://bigsdb.pasteur.fr/api'},
    'bigsdb_auth': {'pubmlst': 'https://rest.pubmlst.org', 'pasteur': 'https://bigsdb.pasteur.fr/api'},
}



import subprocess
from pathlib import Path
from typing import Optional

from mist.app.loggers.logger import logger


class Command:
    """
    Class to handle the execution of commands
    """

    def __init__(self, command: str = None) -> None:
        """
        Initializes the command object.
        :param command: (optional) Command line call
        """
        self._stdout = None
        self._stderr = None
        self._procedure = None
        self._exit_code = None
        self._command = command

    @property
    def stderr(self) -> str:
        """
        Returns the stderr from the command execution.
        :return: Standard error
        """
        return self._stderr

    @property
    def stdout(self) -> str:
        """
        Returns the stdout from the command execution.
        :return: Standard error
        """
        return self._stdout

    @property
    def exit_code(self) -> int:
        """
        Returns the exit code from the command execution.
        :return: Exit code
        """
        return self._exit_code

    @property
    def command(self) -> str:
        """
        Returns the command line call.
        :return: Command line call
        """
        return self._command

    @command.setter
    def command(self, cmd: str) -> None:
        """
        Sets the command line call.
        :param cmd: Command
        :return: None
        """
        self._command = cmd

    def run(self, folder: Path, stderr_handle=subprocess.PIPE, disable_logging: bool = False,
            timeout: Optional[int] = None) -> None:
        """
        Runs the command given at command initialization
        :param folder: Folder where the command is executed
        :param stderr_handle: Handle for the standard error (e.g. PIPE or STDOUT)
        :param disable_logging: If True, logging is disabled
        :param timeout: Timeout for the command execution
        :return: None
        """
        if disable_logging is False:
            logger.info(f'Executing command: {self.command}')
        if self.command is None:
            raise ValueError("Invalid command 'None'")
        self._procedure = subprocess.run(
            self._command,
            stdout=subprocess.PIPE,
            stderr=stderr_handle,
            shell=True,
            executable='/bin/bash',
            cwd=folder,
            timeout=timeout,
            text=True)
        self._stdout = self._procedure.stdout or ''
        self._stderr = self._procedure.stderr or ''
        self._exit_code = self._procedure.returncode
        if disable_logging is False:
            logger.debug(f'stdout: {self._stdout}')
            logger.debug(f'stderr: {self._stderr}')



import json
import shutil
from pathlib import Path
from typing import Any

from Bio import SeqIO

from mist.app.loggers.logger import logger
from mist.app.utils import nucmerutils


class ClusterSplit:
    """
    This class is used to split clusters into groups of loci that do not overlap (i.e., have the same exact start and
    stop position).
    """

    def __init__(self, dir_in: Path, debug: bool) -> None:
        """
        Initializes this class.
        :param dir_in: Input directory
        :param debug: If True, enables debug mode
        :return: None
        """
        with open(dir_in / 'mist_db.json') as handle:
            self._metadata = json.load(handle)
        with open(dir_in / self._metadata['fasta_full']) as handle:
            self._seq_by_id = {s.id: s for s in SeqIO.parse(handle, 'fasta')}
        with open(dir_in / self._metadata['fasta_clustered']) as handle:
            self._seqs_clust = list(SeqIO.parse(handle, 'fasta'))
        self._dir_in = dir_in
        self._dir_temp = dir_in / 'tmp'
        self._dir_temp.mkdir(exist_ok=True)
        self._locus = self._metadata['name']
        self._debug = debug

    def run(self) -> None:
        """
        Runs the cluster splitting.
        :return: None
        """
        for _, cluster_data in self._metadata['clusters'].items():
            if len(cluster_data['members']) <= 1:
                continue
            self._seqs_clust.extend(self.process_cluster(cluster_data))
        path_out_fasta_clust_ext = self._dir_in / self._metadata['fasta_clustered']
        with open(path_out_fasta_clust_ext, 'w') as handle:
            SeqIO.write(self._seqs_clust, handle, 'fasta')
        shutil.rmtree(self._dir_temp)

    def process_cluster(self, cluster_data: dict[str, Any]) -> list[SeqIO.SeqRecord]:
        """
        Processes a cluster and returns the additional sequences that should be included in the clustered FASTA file.
        :param cluster_data: Cluster data
        :return: List of sequences to add
        """
        cluster_name = cluster_data['name']

        # Create FASTA with all sequences from the cluster
        path_fasta = self.export_seqs_from_cluster(cluster_data)
        path_fasta_ref = self.export_seqs_from_cluster(cluster_data, only_first=True)

        # Run nucmer
        path_nucmer = nucmerutils.nucmer(path_fasta_ref, path_fasta, self._dir_temp, debug=self._debug)
        data_coords = nucmerutils.show_coords(path_nucmer, debug=self._debug)

        # Calculate the offset
        data_coords['offset'] = data_coords['[S2]'] - data_coords['[S1]']
        if len(data_coords['offset'].unique()) == 1:
            return []
        seq_ids = [d.iloc[0]['[TAGS]'] for offset, d in data_coords.groupby('offset') if offset != 0]
        logger.debug(f"Splitting cluster {cluster_name} ({self._locus}) into {len(seq_ids) + 1} groups")
        logger.debug(f"Novel representative(s): {seq_ids}")
        return [self._seq_by_id[seq_id] for seq_id in seq_ids]

    def export_seqs_from_cluster(self, cluster_data: dict[str, Any], only_first: bool = False) -> Path:
        """
        Export the sequences from the given cluster.
        :param cluster_data: Cluster data
        :param only_first: If True, only the first is selected
        :return: FASTA file with sequences
        """
        # Create output path
        if only_first:
            basename = f"_repr-{cluster_data['name']}"
        else:
            basename = f"_all-{cluster_data['name']}"
        path_out = Path(self._dir_temp, f'{basename}.fasta')

        # Export the FASTA file
        with path_out.open('w') as handle:
            if only_first:
                SeqIO.write(self._seq_by_id[cluster_data['members'][0]['seq_id']], handle, 'fasta')
            else:
                SeqIO.write([self._seq_by_id[member['seq_id']] for member in cluster_data['members']], handle, 'fasta')
        return path_out



import tempfile


def get_temp_dir() -> tempfile.TemporaryDirectory:
    """
    Returns a temporary directory for testing.
    :return: Temporary directory
    """
    return tempfile.TemporaryDirectory()



import re
import unittest

from click.testing import CliRunner

from mist.app.loggers.logger import initialize_logging
from mist.scripts.cli import cli


class TestList(unittest.TestCase):
    """
    Tests for the listing functionality.
    """

    @staticmethod
    def _parse_stdout(stdout: str) -> list[tuple[str, str]]:
        """
        Parses the schemes from the stdout.
        :param stdout: Tool stdout
        :return: List of schemes and URLs
        """
        schemes = []
        for row in stdout.splitlines():
            if row.startswith('----'):
                break
            scheme, second = row.rsplit(maxsplit=1)
            schemes.append((scheme, second))
        return schemes

    def test_list_cgmlst_org(self) -> None:
        """
        Tests listing the schemes from cgMLST.org.
        :return: None
        """
        runner = CliRunner()
        # noinspection PyTypeChecker
        result = runner.invoke(
            cli,[
                'list',
                '--downloader', 'cgmlstorg'
            ], catch_exceptions=False
        )
        schemes = TestList._parse_stdout(result.output)
        self.assertGreater(len(schemes), 20, "Fewer schemes listed than expected")
        self.assertTrue(result.exit_code == 0)

    def test_list_enterobase(self) -> None:
        """
        Tests listing the schemes from EnteroBase.
        :return: None
        """
        runner = CliRunner()
        # noinspection PyTypeChecker
        result = runner.invoke(
            cli,[
                'list',
                '--downloader', 'enterobase'
            ], catch_exceptions=False
        )
        schemes = TestList._parse_stdout(result.output)
        self.assertGreater(len(schemes), 10, "Fewer schemes listed than expected")
        self.assertTrue(result.exit_code == 0)

    def test_list_bigsdb_pubmlst_species(self) -> None:
        """
        Tests listing the species from PubMLST.
        :return: None
        """
        runner = CliRunner()
        # noinspection PyTypeChecker
        result = runner.invoke(
            cli,[
                'list',
                '--downloader', 'bigsdb',
                '--host', 'pubmlst'
            ], catch_exceptions=False
        )
        schemes = TestList._parse_stdout(result.output)
        self.assertGreater(len(schemes), 50, "Fewer schemes listed than expected")
        self.assertTrue(result.exit_code == 0)

    def test_list_bigsdb_pasteur_species(self) -> None:
        """
        Tests listing the species from BIGSdb Pasteur.
        :return: None
        """
        runner = CliRunner()
        # noinspection PyTypeChecker
        result = runner.invoke(
            cli,[
                'list',
                '--downloader', 'bigsdb',
                '--host', 'pubmlst'
            ], catch_exceptions=False
        )
        schemes = TestList._parse_stdout(result.output)
        self.assertGreater(len(schemes), 10, "Fewer schemes listed than expected")
        self.assertTrue(result.exit_code == 0)

    def test_list_bigsdb_pubmlst_schemes(self) -> None:
        """
        Tests listing the schemes from BIGSdb from a particular species.
        :return: None
        """
        runner = CliRunner()

        # Retrieve the species
        # noinspection PyTypeChecker
        result = runner.invoke(
            cli,[
                'list',
                '--downloader', 'bigsdb',
                '--host', 'pubmlst',
            ], catch_exceptions=False)
        species = TestList._parse_stdout(result.output)

        # Retrieve the DB name for Neisseria
        neisseria_db_name = next(url for name, url in species if 'neisseria' in name.lower())

        # noinspection PyTypeChecker
        result = runner.invoke(
            cli,
            [
                'list',
                '--downloader', 'bigsdb',
                '--host', 'pubmlst',
                '--db', neisseria_db_name
            ], catch_exceptions=False)
        schemes = TestList._parse_stdout(result.output)
        self.assertGreater(len(schemes), 20, "Fewer schemes listed than expected")
        self.assertTrue(result.exit_code == 0)

    def test_list_bigsdb_pasteur_schemes(self) -> None:
        """
        Tests listing the schemes from BIGSdb from a particular species.
        :return: None
        """
        runner = CliRunner()

        # Retrieve the species
        # noinspection PyTypeChecker
        result = runner.invoke(
            cli,[
                'list',
                '--downloader', 'bigsdb',
                '--host', 'pasteur',
            ], catch_exceptions=False)
        species = TestList._parse_stdout(result.output)

        # Retrieve the DB name for Neisseria
        db_name = next(url for name, url in species if 'klebsiella' in name.lower())

        # noinspection PyTypeChecker
        result = runner.invoke(
            cli,
            [
                'list',
                '--downloader', 'bigsdb',
                '--host', 'pasteur',
                '--db', db_name
            ], catch_exceptions=False)
        schemes = TestList._parse_stdout(result.output)
        self.assertGreater(len(schemes), 20, "Fewer schemes listed than expected")
        self.assertTrue(result.exit_code == 0)


if __name__ == '__main__':
    initialize_logging()
    unittest.main()



import gzip
import re
import shutil
from pathlib import Path
from typing import Any

import bs4
import requests
from furl import furl

from mist.app.dbs.basedownloader import BaseDownloader
from mist.app.loggers.logger import logger


class EnteroBaseDownloader(BaseDownloader):
    """
    Downloads (cg)MLST schemes from EnteroBase.
    """

    DOWNLOADER_KEY = 'enterobase'
    STATUS_CODE_OK = 200

    @staticmethod
    def extract_fasta_gz_urls(url: furl) -> list[furl]:
        """
        Extracts all `.fasta.gz` file URLs from an HTML directory listing.
        :param url: Base directory URL (as `furl`)
        :return: List of full URLs (as `furl`) to `.fasta.gz` files
        """
        response = requests.get(url.tostr())
        if response.status_code != EnteroBaseDownloader.STATUS_CODE_OK:
            raise RuntimeError(f'Error accessing: {url} (status code: {response.status_code})')
        content = bs4.BeautifulSoup(response.text, 'html.parser')
        links = content.find_all('a')
        return [
            url.copy().add(path=a['href'])
            for a in links
            if a['href'].endswith('.fasta.gz')
        ]

    @staticmethod
    def download_fasta_file(dir_out: Path, url: furl) -> Path:
        """
        Downloads the given FASTA file.
        :param dir_out: Directory to store the FASTA file
        :param url: URL to the FASTA file
        :return: Path to FASTA file
        """
        logger.debug(f"Downloading FASTA file: {url.path.segments[-1]}")
        locus_name = re.search(r'(.+)\.fasta\.gz$', url.path.segments[-1]).group(1)
        path_out = dir_out / f'{locus_name}.fasta'

        # Download and decompress
        response = requests.get(url.tostr(), stream=True)
        if response.status_code != EnteroBaseDownloader.STATUS_CODE_OK:
            raise RuntimeError(f'Error accessing: {url} (status code: {response.status_code})')
        with (
            gzip.GzipFile(fileobj=response.raw) as gzip_file,
            open(path_out, 'wb') as handle_out,
        ):
            shutil.copyfileobj(gzip_file, handle_out)
        logger.debug(f'FASTA file downloaded: {path_out.name}')
        return path_out

    @staticmethod
    def download_profiles(dir_out: Path, url: furl) -> Path:
        """
        Downloads the profiles file.
        :param dir_out: Output directory
        :param url: Scheme url
        :return: Path to profiles file
        """
        href_profiles = url.copy().add(path='profiles.list.gz')
        path_out = dir_out / 'profiles.tsv'

        # Download and decompress
        response = requests.get(href_profiles.tostr(), stream=True)
        with (
            gzip.GzipFile(fileobj=response.raw) as gzip_file,
            open(path_out, 'wb') as handle_out,
        ):
            shutil.copyfileobj(gzip_file, handle_out)
        logger.info(f'Profiles file downloaded: {path_out.name}')
        return path_out

    def _download(self, url: str, dir_out: Path, include_profiles: bool = False) -> None:
        """
        Downloads the target scheme.
        :param url: Scheme URL
        :param dir_out: Output directory
        :param include_profiles: Include profiles
        :return: None
        """
        # Get the URLs to the FASTA files
        urls_fasta = EnteroBaseDownloader.extract_fasta_gz_urls(furl(url))
        logger.info(f'Found {len(urls_fasta):,} FASTA files')
        if len(urls_fasta) == 0:
            raise RuntimeError('No FASTA file URLs extracted')

        # Download the FASTA files
        paths_fasta = []
        for url_fasta in urls_fasta:
            paths_fasta.append(EnteroBaseDownloader.download_fasta_file(dir_out, url_fasta))
        self.create_fasta_list(paths_fasta)

        # Download the profiles file
        if include_profiles:
            EnteroBaseDownloader.download_profiles(dir_out, furl(url))
        logger.info(f'Scheme downloaded to: {dir_out}')

    def get_available_schemes(self, base_url: furl, **kwargs: Any) -> list[tuple[str, str]]:
        """
        Retrieve a list of available cgMLST schemes and their URL.
        :param base_url: Base URL
        :param kwargs: Keyword arguments
        :return: None
        """
        response = requests.get(str(base_url))
        soup = bs4.BeautifulSoup(response.text, 'html.parser')
        schemes = []
        for a in soup.select('pre a'):
            href = a.get('href')
            name = a.get_text(strip=True).rstrip('/')

            # Skip parent directory entry
            if href == '../':
                continue
            url = base_url.copy().add(path=href)
            schemes.append((name, str(url)))
        return schemes



from collections.abc import Sequence


def unique_preserve_order(seq: Sequence) -> list:
    """
    Only keeps the unique elements from the input list while preserving the order.
    :param seq: Input sequence
    :return: List of unique elements
    """
    seen = set()
    result = []
    for item in seq:
        if item in seen:
            continue
        seen.add(item)
        result.append(item)
    return result
